import base64
import dataclasses
import logging
import mimetypes
import random
import re
import time
from abc import ABC, abstractmethod
from datetime import datetime
from ftplib import FTP
from io import BytesIO
from pathlib import Path
from posixpath import split
from shutil import rmtree
from tempfile import mkdtemp
from typing import Any
from unittest import mock
from unittest.mock import MagicMock

import attr
import pytest
from itemadapter import ItemAdapter
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

from scrapy.crawler import Crawler
from scrapy.exceptions import IgnoreRequest, NotConfigured, ScrapyDeprecationWarning
from scrapy.http import Request, Response
from scrapy.item import Field, Item
from scrapy.pipelines import files
from scrapy.pipelines.files import (
    FilesPipeline,
    FSFilesStore,
    FTPFilesStore,
    GCSFilesStore,
    S3FilesStore,
)
from scrapy.pipelines.media import FileException, _MediaRequestFiltered
from scrapy.settings import Settings
from scrapy.utils.asyncio import call_later
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.spider import DefaultSpider
from scrapy.utils.test import get_crawler
from tests.mockserver.ftp import MockFTPServer
from tests.utils.decorators import coroutine_test, inline_callbacks_test

from .utils.cloud import mock_google_cloud_storage
from .utils.media_pipelines import DUMMY_SPIDER_INFO, mocked_download_func


def get_ftp_content_and_delete(
    path: str,
    host: str,
    port: int,
    username: str,
    password: str,
    use_active_mode: bool = False,
) -> bytes:
    with FTP() as ftp:
        ftp.connect(host, port)
        ftp.login(username, password)
        if use_active_mode:
            ftp.set_pasv(False)
        ftp_data: list[bytes] = []

        def buffer_data(data: bytes) -> None:
            ftp_data.append(data)

        ftp.retrbinary(f"RETR {path}", buffer_data)
        dirname, filename = split(path)
        ftp.cwd(dirname)
        ftp.delete(filename)
    return b"".join(ftp_data)


class DeferredFSFilesStore(FSFilesStore):
    """A simple store with persist_file() returning a deferred."""

    def persist_file(self, path, buf, info, meta=None, headers=None):
        deferred: Deferred[None] = Deferred()
        # short-hand super() doesn't work in nested functions
        parent_persist_file = super().persist_file

        def cb():
            parent_persist_file(path, buf, info, meta=meta, headers=headers)
            deferred.callback(None)

        call_later(0.5, cb)
        return deferred


class TestFilesPipeline:
    def setup_method(self):
        self.tempdir = mkdtemp()
        self.pipeline = self._create_pipeline(FilesPipeline)

    def teardown_method(self):
        rmtree(self.tempdir)

    def _create_pipeline(
        self,
        pipeline_cls: type[FilesPipeline],
        settings: dict[str, Any] | None = None,
    ) -> FilesPipeline:
        crawler = get_crawler(
            DefaultSpider, {"FILES_STORE": self.tempdir, **(settings or {})}
        )
        crawler.spider = crawler._create_spider()
        crawler.engine = MagicMock(download_async=mocked_download_func)
        pipeline = build_from_crawler(pipeline_cls, crawler)
        pipeline.open_spider()
        return pipeline

    def test_file_path_query_parameters(self):
        file_path = self.pipeline.file_path

        req1 = Request("http://foo.bar/baz.txt?fizz")
        assert file_path(req1) == "full/a2b4913a62f65445aeae2bac08cd8c3b41d7195e.txt"

        req2 = Request("http://foo.bar/get_img.foo?file=photo.jpg")
        assert file_path(req2) == "full/7fc9461c9fd836515bea6983373097203a7d748e.jpg"

    def test_file_path(self):
        file_path = self.pipeline.file_path
        assert (
            file_path(Request("https://dev.mydeco.com/mydeco.pdf"))
            == "full/c9b564df929f4bc635bdd19fde4f3d4847c757c5.pdf"
        )
        assert (
            file_path(
                Request(
                    "http://www.maddiebrown.co.uk///catalogue-items//image_54642_12175_95307.txt"
                )
            )
            == "full/4ce274dd83db0368bafd7e406f382ae088e39219.txt"
        )
        assert (
            file_path(
                Request("https://dev.mydeco.com/two/dirs/with%20spaces%2Bsigns.doc")
            )
            == "full/94ccc495a17b9ac5d40e3eabf3afcb8c2c9b9e1a.doc"
        )
        assert (
            file_path(
                Request(
                    "http://www.dfsonline.co.uk/get_prod_image?img=status_0907_mdm.jpg"
                )
            )
            == "full/c67f916ff9d542e822dedf38f9fcb146d1faba78.jpg"
        )
        assert (
            file_path(Request("http://www.dorma.co.uk/images/product_details/2532/"))
            == "full/97ee6f8a46cbbb418ea91502fd24176865cf39b2"
        )
        assert (
            file_path(Request("http://www.dorma.co.uk/images/product_details/2532"))
            == "full/244e0dd7d96a3b7b01f54eded250c9e272577aa1"
        )
        assert (
            file_path(
                Request("http://www.dorma.co.uk/images/product_details/2532"),
                response=Response("http://www.dorma.co.uk/images/product_details/2532"),
                info=object(),  # type: ignore[arg-type]
            )
            == "full/244e0dd7d96a3b7b01f54eded250c9e272577aa1"
        )
        assert (
            file_path(
                Request(
                    "http://www.dfsonline.co.uk/get_prod_image?img=status_0907_mdm.jpg.bohaha"
                )
            )
            == "full/e75f2fa260521b56f6b6a867447b8002d00b5841"
        )
        assert (
            file_path(
                Request(
                    "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAR0AAACxCAMAAADOHZloAAACClBMVEX/\
                                    //+F0tzCwMK76ZKQ21AMqr7oAAC96JvD5aWM2kvZ78J0N7fmAAC46Y4Ap7y"
                )
            )
            == "full/178059cbeba2e34120a67f2dc1afc3ecc09b61cb.png"
        )

    def test_fs_store(self):
        assert isinstance(self.pipeline.store, FSFilesStore)
        assert self.pipeline.store.basedir == self.tempdir

        path = "some/image/key.jpg"
        fullpath = Path(self.tempdir, "some", "image", "key.jpg")
        assert self.pipeline.store._get_filesystem_path(path) == fullpath

    @coroutine_test
    async def test_file_not_expired(self):
        item_url = "http://example.com/file.pdf"
        item = _create_item_with_files(item_url)
        with (
            mock.patch.object(FilesPipeline, "inc_stats", return_value=True),
            mock.patch.object(
                FSFilesStore,
                "stat_file",
                return_value={"checksum": "abc", "last_modified": time.time()},
            ),
            mock.patch.object(
                FilesPipeline,
                "get_media_requests",
                return_value=[_prepare_request_object(item_url)],
            ),
        ):
            result = await self.pipeline.process_item(item)
        assert result["files"][0]["checksum"] == "abc"
        assert result["files"][0]["status"] == "uptodate"

    @coroutine_test
    async def test_file_expired(self):
        item_url = "http://example.com/file2.pdf"
        item = _create_item_with_files(item_url)
        with (
            mock.patch.object(
                FSFilesStore,
                "stat_file",
                return_value={
                    "checksum": "abc",
                    "last_modified": time.time()
                    - (self.pipeline.expires * 60 * 60 * 24 * 2),
                },
            ),
            mock.patch.object(
                FilesPipeline,
                "get_media_requests",
                return_value=[_prepare_request_object(item_url)],
            ),
            mock.patch.object(FilesPipeline, "inc_stats", return_value=True),
        ):
            result = await self.pipeline.process_item(item)
        assert result["files"][0]["checksum"] != "abc"
        assert result["files"][0]["status"] == "downloaded"

    @coroutine_test
    async def test_file_cached(self):
        item_url = "http://example.com/file3.pdf"
        item = _create_item_with_files(item_url)
        with (
            mock.patch.object(FilesPipeline, "inc_stats", return_value=True),
            mock.patch.object(
                FSFilesStore,
                "stat_file",
                return_value={
                    "checksum": "abc",
                    "last_modified": time.time()
                    - (self.pipeline.expires * 60 * 60 * 24 * 2),
                },
            ),
            mock.patch.object(
                FilesPipeline,
                "get_media_requests",
                return_value=[_prepare_request_object(item_url, flags=["cached"])],
            ),
        ):
            result = await self.pipeline.process_item(item)
        assert result["files"][0]["checksum"] != "abc"
        assert result["files"][0]["status"] == "cached"

    @coroutine_test
    async def test_file_stat_without_last_modified(self) -> None:
        """A stat result without a last modification time forces a download."""
        item_url = "http://example.com/file4.pdf"
        item = _create_item_with_files(item_url)
        with (
            mock.patch.object(FilesPipeline, "inc_stats", return_value=True),
            mock.patch.object(
                FSFilesStore, "stat_file", return_value={"checksum": "abc"}
            ),
            mock.patch.object(
                FilesPipeline,
                "get_media_requests",
                return_value=[_prepare_request_object(item_url)],
            ),
        ):
            result = await self.pipeline.process_item(item)
        assert result["files"][0]["checksum"] != "abc"
        assert result["files"][0]["status"] == "downloaded"

    @coroutine_test
    async def test_file_empty_content(self, caplog: pytest.LogCaptureFixture) -> None:
        item_url = "http://example.com/empty.pdf"
        item = _create_item_with_files(item_url)
        request = Request(
            item_url, meta={"response": Response(item_url, status=200, body=b"")}
        )
        with (
            caplog.at_level(logging.WARNING),
            mock.patch.object(
                FilesPipeline, "get_media_requests", return_value=[request]
            ),
        ):
            result = await self.pipeline.process_item(item)
        assert result["files"] == []
        assert "File (empty-content): Empty file from" in caplog.text

    @coroutine_test
    async def test_file_downloaded_file_exception(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        """A FileException from file_downloaded() is logged as a warning and
        kept as is."""

        class FailingFilesPipeline(FilesPipeline):
            def file_downloaded(self, response, request, info, *, item=None):
                raise FileException("boom")

        item_url = "http://example.com/file5.pdf"
        item = _create_item_with_files(item_url)
        pipeline = self._create_pipeline(FailingFilesPipeline)
        with (
            caplog.at_level(logging.WARNING),
            mock.patch.object(
                FilesPipeline,
                "get_media_requests",
                return_value=[_prepare_request_object(item_url)],
            ),
        ):
            result = await pipeline.process_item(item)
        assert result["files"] == []
        records = [
            r for r in caplog.records if "Error processing file" in r.getMessage()
        ]
        assert len(records) == 1
        assert records[0].levelname == "WARNING"
        assert "boom" in records[0].getMessage()

    @coroutine_test
    async def test_file_downloaded_unknown_error(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Any other exception from file_downloaded() is logged as an error and
        reported as a FileException."""

        class FailingFilesPipeline(FilesPipeline):
            def file_downloaded(self, response, request, info, *, item=None):
                raise RuntimeError("boom")

        item_url = "http://example.com/file6.pdf"
        item = _create_item_with_files(item_url)
        pipeline = self._create_pipeline(FailingFilesPipeline)
        with (
            caplog.at_level(logging.WARNING),
            mock.patch.object(
                FilesPipeline,
                "get_media_requests",
                return_value=[_prepare_request_object(item_url)],
            ),
        ):
            result = await pipeline.process_item(item)
        assert result["files"] == []
        records = [
            r for r in caplog.records if "Error processing file" in r.getMessage()
        ]
        assert len(records) == 1
        assert records[0].levelname == "ERROR"
        exc_info = records[0].exc_info
        assert exc_info is not None
        assert exc_info[0] is RuntimeError

    @coroutine_test
    async def test_async_store(self) -> None:
        """Test that async persist_file() works and is awaited."""

        self.pipeline.store = DeferredFSFilesStore(self.tempdir)
        item_url = "http://example.com/file.pdf"
        item = _create_item_with_files(item_url)
        with (
            mock.patch.object(FilesPipeline, "inc_stats", return_value=True),
            mock.patch.object(
                FilesPipeline,
                "get_media_requests",
                return_value=[_prepare_request_object(item_url)],
            ),
        ):
            result = await self.pipeline.process_item(item)
        assert result["files"][0]["status"] == "downloaded"
        assert result["files"][0]["checksum"]
        # check that the file was written by persist_file()
        path = Path(self.tempdir) / result["files"][0]["path"]
        assert path.exists()
        assert path.read_bytes() == b"data"

    def test_file_path_from_item(self):
        """
        Custom file path based on item data, overriding default implementation
        """

        class CustomFilesPipeline(FilesPipeline):
            def file_path(self, request, response=None, info=None, item=None) -> str:
                return f"full/{item.get('path')}"

        file_path = build_from_crawler(
            CustomFilesPipeline, get_crawler(None, {"FILES_STORE": self.tempdir})
        ).file_path
        item = {"path": "path-to-store-file"}
        request = Request("http://example.com")
        assert file_path(request, item=item) == "full/path-to-store-file"

    @coroutine_test
    async def test_file_path_from_response(self) -> None:
        """file_path() may build the path out of the response, e.g. to get the
        file extension from a response header, as long as FILES_EXPIRES is 0 to
        disable the up-to-date check, which runs before the download and hence
        cannot reach the same path."""

        class ContentTypeFilesPipeline(FilesPipeline):
            def file_path(self, request, response=None, info=None, *, item=None):
                path = super().file_path(request, response, info, item=item)
                if response is None:
                    return path
                content_type = response.headers["Content-Type"].decode()
                return path + (mimetypes.guess_extension(content_type) or "")

        item_url = "http://example.com/download?id=1"
        item = _create_item_with_files(item_url)
        pipeline = self._create_pipeline(ContentTypeFilesPipeline, {"FILES_EXPIRES": 0})
        request = _prepare_request_object(
            item_url, headers={"Content-Type": "application/pdf"}
        )
        with (
            mock.patch.object(FilesPipeline, "inc_stats", return_value=True),
            # A fresh file at the response-less path is ignored thanks to
            # FILES_EXPIRES being 0.
            mock.patch.object(
                FSFilesStore,
                "stat_file",
                return_value={"checksum": "abc", "last_modified": time.time() - 1},
            ),
            mock.patch.object(
                FilesPipeline, "get_media_requests", return_value=[request]
            ),
        ):
            result = await pipeline.process_item(item)

        file_info = result["files"][0]
        assert file_info["status"] == "downloaded"
        assert file_info["path"].endswith(".pdf")
        assert (Path(self.tempdir) / file_info["path"]).read_bytes() == b"data"

    def test_media_failed_filtered_request(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        """A filtered media request (IgnoreRequest) is reported as a
        _MediaRequestFiltered exception and logged at the DEBUG level, instead
        of as a download error with a traceback."""
        request = Request("http://example.com/file.pdf")
        reason = "Filtered offsite request to 'example.com'"
        failure = Failure(IgnoreRequest(reason))

        with (
            caplog.at_level(logging.DEBUG),
            pytest.raises(_MediaRequestFiltered, match=re.escape(reason)),
        ):
            self.pipeline.media_failed(failure, request, self.pipeline.spiderinfo)

        assert len(caplog.records) == 1
        record = caplog.records[0]
        assert record.levelname == "DEBUG"
        assert record.exc_info is None
        assert reason in record.getMessage()

    def test_media_failed_download_error(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        """A genuine download error is reported as a FileException and logged as
        a warning."""
        request = Request("http://example.com/file.pdf")
        failure = Failure(Exception("boom"))

        with caplog.at_level(logging.WARNING), pytest.raises(FileException):
            self.pipeline.media_failed(failure, request, self.pipeline.spiderinfo)

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == "WARNING"

    @coroutine_test
    async def test_process_item_filtered_request(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        """A filtered (e.g. offsite) media request is processed as a failed
        result without being logged as an error with a traceback."""
        item_url = "http://example.com/file.pdf"
        item = _create_item_with_files(item_url)
        request = Request(
            item_url,
            meta={
                "response": IgnoreRequest("Filtered offsite request to 'example.com'")
            },
        )
        with (
            caplog.at_level(logging.DEBUG),
            mock.patch.object(
                FilesPipeline, "get_media_requests", return_value=[request]
            ),
        ):
            result = await self.pipeline.process_item(item)

        assert result["files"] == []
        assert not any(r.levelname in ("WARNING", "ERROR") for r in caplog.records)
        assert any(
            "Filtered offsite request to 'example.com'" in r.getMessage()
            for r in caplog.records
        )

    @pytest.mark.parametrize(
        "bad_type",
        [
            "http://example.com/file.pdf",
            ("http://example.com/file.pdf",),
            {"url": "http://example.com/file.pdf"},
            123,
            None,
        ],
    )
    def test_rejects_non_list_file_urls(self, tmp_path, bad_type):
        pipeline = build_from_crawler(
            FilesPipeline, get_crawler(None, {"FILES_STORE": str(tmp_path)})
        )
        item = ItemWithFiles()
        item["file_urls"] = bad_type

        with pytest.raises(TypeError, match="file_urls must be a list of URLs"):
            list(pipeline.get_media_requests(item, None))  # type: ignore[arg-type]


class TestFilesPipelineFieldsMixin(ABC):
    @property
    @abstractmethod
    def item_class(self) -> Any:
        raise NotImplementedError

    def test_item_fields_default(self, tmp_path):
        url = "http://www.example.com/files/1.txt"
        item = self.item_class(name="item1", file_urls=[url])
        pipeline = build_from_crawler(
            FilesPipeline, get_crawler(None, {"FILES_STORE": tmp_path})
        )
        requests = list(pipeline.get_media_requests(item, None))  # type: ignore[arg-type]
        assert requests[0].url == url
        results = [(True, {"url": url})]
        item = pipeline.item_completed(results, item, None)  # type: ignore[arg-type]
        files = ItemAdapter(item).get("files")
        assert files == [results[0][1]]
        assert isinstance(item, self.item_class)

    def test_item_fields_override_settings(self, tmp_path):
        url = "http://www.example.com/files/1.txt"
        item = self.item_class(name="item1", custom_file_urls=[url])
        pipeline = build_from_crawler(
            FilesPipeline,
            get_crawler(
                None,
                {
                    "FILES_STORE": tmp_path,
                    "FILES_URLS_FIELD": "custom_file_urls",
                    "FILES_RESULT_FIELD": "custom_files",
                },
            ),
        )
        requests = list(pipeline.get_media_requests(item, None))  # type: ignore[arg-type]
        assert requests[0].url == url
        results = [(True, {"url": url})]
        item = pipeline.item_completed(results, item, None)  # type: ignore[arg-type]
        custom_files = ItemAdapter(item).get("custom_files")
        assert custom_files == [results[0][1]]
        assert isinstance(item, self.item_class)


class TestFilesPipelineFieldsDict(TestFilesPipelineFieldsMixin):
    item_class = dict


class FilesPipelineTestItem(Item):
    name = Field()
    # default fields
    file_urls = Field()
    files = Field()
    # overridden fields
    custom_file_urls = Field()
    custom_files = Field()


class TestFilesPipelineFieldsItem(TestFilesPipelineFieldsMixin):
    item_class = FilesPipelineTestItem


@dataclasses.dataclass
class FilesPipelineTestDataClass:
    name: str
    # default fields
    file_urls: list[str] = dataclasses.field(default_factory=list)
    files: list[dict[str, str]] = dataclasses.field(default_factory=list)
    # overridden fields
    custom_file_urls: list[str] = dataclasses.field(default_factory=list)
    custom_files: list[dict[str, str]] = dataclasses.field(default_factory=list)


class TestFilesPipelineFieldsDataClass(TestFilesPipelineFieldsMixin):
    item_class = FilesPipelineTestDataClass


@attr.s
class FilesPipelineTestAttrsItem:
    name = attr.ib(default="")
    # default fields
    file_urls: list[str] = attr.ib(default=list)
    files: list[dict[str, str]] = attr.ib(default=list)
    # overridden fields
    custom_file_urls: list[str] = attr.ib(default=list)
    custom_files: list[dict[str, str]] = attr.ib(default=list)


class TestFilesPipelineFieldsAttrsItem(TestFilesPipelineFieldsMixin):
    item_class = FilesPipelineTestAttrsItem


class TestFilesPipelineCustomSettings:
    default_cls_settings = {
        "EXPIRES": 90,
        "FILES_URLS_FIELD": "file_urls",
        "FILES_RESULT_FIELD": "files",
    }
    file_cls_attr_settings_map = {
        ("EXPIRES", "FILES_EXPIRES", "expires"),
        ("FILES_URLS_FIELD", "FILES_URLS_FIELD", "files_urls_field"),
        ("FILES_RESULT_FIELD", "FILES_RESULT_FIELD", "files_result_field"),
    }

    def _generate_fake_settings(
        self, tmp_path: Path, prefix: str | None = None
    ) -> dict[str, Any]:
        def random_string() -> str:
            return "".join([chr(random.randint(97, 123)) for _ in range(10)])

        settings = {
            "FILES_EXPIRES": random.randint(100, 1000),
            "FILES_URLS_FIELD": random_string(),
            "FILES_RESULT_FIELD": random_string(),
            "FILES_STORE": tmp_path,
        }
        if not prefix:
            return settings

        return {
            prefix.upper() + "_" + k if k != "FILES_STORE" else k: v
            for k, v in settings.items()
        }

    def _generate_fake_pipeline(self) -> type[FilesPipeline]:
        class UserDefinedFilePipeline(FilesPipeline):
            EXPIRES = 1001
            FILES_URLS_FIELD = "alfa"
            FILES_RESULT_FIELD = "beta"

        return UserDefinedFilePipeline

    def test_different_settings_for_different_instances(self, tmp_path):
        """
        If there are different instances with different settings they should keep
        different settings.
        """
        custom_settings = self._generate_fake_settings(tmp_path)
        another_pipeline = build_from_crawler(
            FilesPipeline, get_crawler(None, custom_settings)
        )
        one_pipeline = build_from_crawler(
            FilesPipeline, get_crawler(None, {"FILES_STORE": tmp_path})
        )
        for pipe_attr, settings_attr, pipe_ins_attr in self.file_cls_attr_settings_map:
            default_value = self.default_cls_settings[pipe_attr]
            assert getattr(one_pipeline, pipe_attr) == default_value
            custom_value = custom_settings[settings_attr]
            assert default_value != custom_value
            assert getattr(another_pipeline, pipe_ins_attr) == custom_value

    def test_subclass_attributes_preserved_if_no_settings(self, tmp_path):
        """
        If subclasses override class attributes and there are no special settings those values should be kept.
        """
        pipe_cls = self._generate_fake_pipeline()
        pipe = build_from_crawler(
            pipe_cls, get_crawler(None, {"FILES_STORE": tmp_path})
        )
        for pipe_attr, _, pipe_ins_attr in self.file_cls_attr_settings_map:
            custom_value = getattr(pipe, pipe_ins_attr)
            assert custom_value != self.default_cls_settings[pipe_attr]
            assert getattr(pipe, pipe_ins_attr) == getattr(pipe, pipe_attr)

    def test_subclass_attrs_preserved_custom_settings(self, tmp_path):
        """
        If file settings are defined but they are not defined for subclass
        settings should be preserved.
        """
        pipeline_cls = self._generate_fake_pipeline()
        settings = self._generate_fake_settings(tmp_path)
        pipeline = build_from_crawler(pipeline_cls, get_crawler(None, settings))
        for pipe_attr, settings_attr, pipe_ins_attr in self.file_cls_attr_settings_map:
            value = getattr(pipeline, pipe_ins_attr)
            setting_value = settings.get(settings_attr)
            assert value != self.default_cls_settings[pipe_attr]
            assert value == setting_value

    def test_no_custom_settings_for_subclasses(self, tmp_path):
        """
        If there are no settings for subclass and no subclass attributes, pipeline should use
        attributes of base class.
        """

        class UserDefinedFilesPipeline(FilesPipeline):
            pass

        user_pipeline = build_from_crawler(
            UserDefinedFilesPipeline, get_crawler(None, {"FILES_STORE": tmp_path})
        )
        for pipe_attr, _, pipe_ins_attr in self.file_cls_attr_settings_map:
            # Values from settings for custom pipeline should be set on pipeline instance.
            custom_value = self.default_cls_settings.get(pipe_attr.upper())
            assert getattr(user_pipeline, pipe_ins_attr) == custom_value

    def test_custom_settings_for_subclasses(self, tmp_path):
        """
        If there are custom settings for subclass and NO class attributes, pipeline should use custom
        settings.
        """

        class UserDefinedFilesPipeline(FilesPipeline):
            pass

        prefix = UserDefinedFilesPipeline.__name__.upper()
        settings = self._generate_fake_settings(tmp_path, prefix=prefix)
        user_pipeline = build_from_crawler(
            UserDefinedFilesPipeline, get_crawler(None, settings)
        )
        for pipe_attr, settings_attr, pipe_inst_attr in self.file_cls_attr_settings_map:
            # Values from settings for custom pipeline should be set on pipeline instance.
            custom_value = settings.get(prefix + "_" + settings_attr)
            assert custom_value != self.default_cls_settings[pipe_attr]
            assert getattr(user_pipeline, pipe_inst_attr) == custom_value

    def test_custom_settings_and_class_attrs_for_subclasses(self, tmp_path):
        """
        If there are custom settings for subclass AND class attributes
        setting keys are preferred and override attributes.
        """
        pipeline_cls = self._generate_fake_pipeline()
        prefix = pipeline_cls.__name__.upper()
        settings = self._generate_fake_settings(tmp_path, prefix=prefix)
        user_pipeline = build_from_crawler(pipeline_cls, get_crawler(None, settings))
        for (
            pipe_cls_attr,
            settings_attr,
            pipe_inst_attr,
        ) in self.file_cls_attr_settings_map:
            custom_value = settings.get(prefix + "_" + settings_attr)
            assert custom_value != self.default_cls_settings[pipe_cls_attr]
            assert getattr(user_pipeline, pipe_inst_attr) == custom_value

    def test_cls_attrs_with_DEFAULT_prefix(self, tmp_path):
        class UserDefinedFilesPipeline(FilesPipeline):
            DEFAULT_FILES_RESULT_FIELD = "this"
            DEFAULT_FILES_URLS_FIELD = "that"

        pipeline = build_from_crawler(
            UserDefinedFilesPipeline, get_crawler(None, {"FILES_STORE": tmp_path})
        )
        assert (
            pipeline.files_result_field
            == UserDefinedFilesPipeline.DEFAULT_FILES_RESULT_FIELD
        )
        assert (
            pipeline.files_urls_field
            == UserDefinedFilesPipeline.DEFAULT_FILES_URLS_FIELD
        )

    def test_user_defined_subclass_default_key_names(self, tmp_path):
        """Test situation when user defines subclass of FilesPipeline,
        but uses attribute names for default pipeline (without prefixing
        them with pipeline class name).
        """
        settings = self._generate_fake_settings(tmp_path)

        class UserPipe(FilesPipeline):
            pass

        pipeline_cls = build_from_crawler(UserPipe, get_crawler(None, settings))

        for _, settings_attr, pipe_inst_attr in self.file_cls_attr_settings_map:
            expected_value = settings.get(settings_attr)
            assert getattr(pipeline_cls, pipe_inst_attr) == expected_value

    def test_file_pipeline_using_pathlike_objects(self, tmp_path):
        class CustomFilesPipelineWithPathLikeDir(FilesPipeline):
            def file_path(self, request, response=None, info=None, *, item=None) -> str:
                return str(Path("subdir") / Path(request.url).name)

        pipeline = build_from_crawler(
            CustomFilesPipelineWithPathLikeDir,
            get_crawler(None, {"FILES_STORE": tmp_path}),
        )
        request = Request("http://example.com/image01.jpg")
        assert pipeline.file_path(request) == str(Path("subdir/image01.jpg"))


class TestFSFilesStore:
    def test_constructor_with_pathlike_object(self, tmp_path: Path) -> None:
        assert FSFilesStore(tmp_path).basedir == str(tmp_path)

    def test_constructor_with_uri(self, tmp_path: Path) -> None:
        assert FSFilesStore(f"file://{tmp_path}").basedir == str(tmp_path)

    def test_stat_file(self, tmp_path: Path) -> None:
        store = FSFilesStore(tmp_path)
        store.persist_file("full/filename", BytesIO(b"data"), DUMMY_SPIDER_INFO)
        stat = store.stat_file("full/filename", DUMMY_SPIDER_INFO)
        assert stat["checksum"] == "8d777f385d3dfec8815d20f7496026dc"
        assert stat["last_modified"] == pytest.approx(time.time(), abs=60)

    def test_stat_missing_file(self, tmp_path: Path) -> None:
        store = FSFilesStore(tmp_path)
        assert store.stat_file("full/filename", DUMMY_SPIDER_INFO) == {}


@pytest.mark.requires_botocore
class TestS3FilesStore:
    @inline_callbacks_test
    def test_persist(self):
        bucket = "mybucket"
        key = "export.csv"
        uri = f"s3://{bucket}/{key}"
        buffer = mock.MagicMock()
        meta = {"foo": "bar"}
        path = ""
        content_type = "image/png"

        store = S3FilesStore(uri)
        from botocore.stub import Stubber  # noqa: PLC0415

        with Stubber(store.s3_client) as stub:
            stub.add_response(
                "put_object",
                expected_params={
                    "ACL": S3FilesStore.POLICY,
                    "Body": buffer,
                    "Bucket": bucket,
                    "CacheControl": S3FilesStore.HEADERS["Cache-Control"],
                    "ContentType": content_type,
                    "Key": key,
                    "Metadata": meta,
                },
                service_response={},
            )

            yield store.persist_file(
                path,
                buffer,
                info=DUMMY_SPIDER_INFO,
                meta=meta,
                headers={"Content-Type": content_type},
            )

            stub.assert_no_pending_responses()
            # The call to read does not happen with Stubber
            assert buffer.method_calls == [mock.call.seek(0)]

    @inline_callbacks_test
    def test_persist_without_headers(self):
        """Without custom headers only the default ones are sent."""
        bucket = "mybucket"
        key = "export.csv"
        buffer = mock.MagicMock()

        store = S3FilesStore(f"s3://{bucket}/{key}")
        from botocore.stub import Stubber  # noqa: PLC0415

        with Stubber(store.s3_client) as stub:
            stub.add_response(
                "put_object",
                expected_params={
                    "ACL": S3FilesStore.POLICY,
                    "Body": buffer,
                    "Bucket": bucket,
                    "CacheControl": S3FilesStore.HEADERS["Cache-Control"],
                    "Key": key,
                    "Metadata": {},
                },
                service_response={},
            )

            yield store.persist_file("", buffer, info=DUMMY_SPIDER_INFO)

            stub.assert_no_pending_responses()

    def test_missing_botocore(self):
        with (
            mock.patch(
                "scrapy.pipelines.files.is_botocore_available", return_value=False
            ),
            pytest.raises(NotConfigured, match="missing botocore library"),
        ):
            S3FilesStore("s3://mybucket/key")

    def test_wrong_uri_scheme(self):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Incorrect URI scheme in ftp://mybucket/key, expected 's3'"
            ),
        ):
            S3FilesStore("ftp://mybucket/key")

    def test_unsupported_header(self):
        store = S3FilesStore("s3://mybucket/key")
        with pytest.raises(
            TypeError, match='Header "X-Custom" is not supported by botocore'
        ):
            store._headers_to_botocore_kwargs({"X-Custom": "value"})

    @inline_callbacks_test
    def test_stat(self):
        bucket = "mybucket"
        key = "export.csv"
        uri = f"s3://{bucket}/{key}"
        checksum = "3187896a9657a28163abb31667df64c8"
        last_modified = datetime(2019, 12, 1)

        store = S3FilesStore(uri)
        from botocore.stub import Stubber  # noqa: PLC0415

        with Stubber(store.s3_client) as stub:
            stub.add_response(
                "head_object",
                expected_params={
                    "Bucket": bucket,
                    "Key": key,
                },
                service_response={
                    "ETag": f'"{checksum}"',
                    "LastModified": last_modified,
                },
            )

            file_stats = yield store.stat_file("", info=DUMMY_SPIDER_INFO)
            assert file_stats == {
                "checksum": checksum,
                "last_modified": last_modified.timestamp(),
            }

            stub.assert_no_pending_responses()

    def test_default_max_pool_connections(self) -> None:
        store = S3FilesStore("s3://mybucket/prefix/")
        config: Any = store.s3_client.meta.config
        assert config.max_pool_connections == 10

    @pytest.mark.parametrize(
        ("settings", "expected"),
        [
            ({}, 10),
            ({"REACTOR_THREADPOOL_MAXSIZE": 20}, 20),
            ({"AWS_MAX_POOL_CONNECTIONS": 30}, 30),
            ({"AWS_MAX_POOL_CONNECTIONS": 30, "REACTOR_THREADPOOL_MAXSIZE": 20}, 30),
        ],
    )
    def test_max_pool_connections(
        self, monkeypatch: pytest.MonkeyPatch, settings: dict[str, Any], expected: int
    ) -> None:
        # restores the value that FilesPipeline.from_crawler() sets on the class
        monkeypatch.setattr(S3FilesStore, "AWS_MAX_POOL_CONNECTIONS", None)
        crawler = get_crawler(
            settings_dict={"FILES_STORE": "s3://mybucket/prefix/", **settings}
        )
        store = build_from_crawler(FilesPipeline, crawler).store
        assert isinstance(store, S3FilesStore)
        config: Any = store.s3_client.meta.config
        assert config.max_pool_connections == expected


class TestGCSFilesStore:
    @staticmethod
    def build_gcs_files_store(
        *,
        permissions: tuple[str, ...] = (
            "storage.objects.get",
            "storage.objects.create",
        ),
    ) -> tuple[GCSFilesStore, Any, Any]:
        """Build a :class:`GCSFilesStore` mock.

        Returns ``(store, bucket_mock, blob_mock)``. Skips the test if
        google-cloud-storage is not installed. ``permissions`` is what
        ``Bucket.test_iam_permissions`` will return.
        """
        pytest.importorskip("google.cloud.storage")

        client_mock, bucket_mock, blob_mock = mock_google_cloud_storage()
        bucket_mock.test_iam_permissions.return_value = list(permissions)

        with mock.patch("google.cloud.storage.Client", return_value=client_mock):
            store = GCSFilesStore("gs://my_bucket/my_prefix/")
        return store, bucket_mock, blob_mock

    def test_init(self, caplog: pytest.LogCaptureFixture) -> None:
        with caplog.at_level(logging.WARNING):
            self.build_gcs_files_store()
        assert not caplog.records

    def test_get_perm_missing(self, caplog: pytest.LogCaptureFixture) -> None:
        self.build_gcs_files_store(permissions=("storage.objects.create",))
        assert (
            "No 'storage.objects.get' permission for GCS bucket my_bucket"
            in caplog.text
        )

    def test_create_perm_missing(self, caplog: pytest.LogCaptureFixture) -> None:
        self.build_gcs_files_store(permissions=("storage.objects.get",))
        assert (
            "No 'storage.objects.create' permission for GCS bucket my_bucket"
            in caplog.text
        )

    def test_update_stores(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setattr(GCSFilesStore, "GCS_PROJECT_ID", None)
        monkeypatch.setattr(GCSFilesStore, "POLICY", None)

        settings = Settings(
            {"GCS_PROJECT_ID": "my-project", "FILES_STORE_GCS_ACL": "publicRead"}
        )
        FilesPipeline._update_stores(settings)
        assert GCSFilesStore.GCS_PROJECT_ID == "my-project"
        assert GCSFilesStore.POLICY == "publicRead"

        # An empty FILES_STORE_GCS_ACL is normalised to None.
        settings = Settings({"GCS_PROJECT_ID": "my-project", "FILES_STORE_GCS_ACL": ""})
        FilesPipeline._update_stores(settings)
        assert GCSFilesStore.POLICY is None

    @coroutine_test
    async def test_persist(self) -> None:
        store, bucket, blob = self.build_gcs_files_store()
        await maybe_deferred_to_future(
            store.persist_file(
                "full/filename",
                BytesIO(b"hello"),
                info=DUMMY_SPIDER_INFO,
                meta={"foo": 1},
                headers={"Content-Type": "image/png"},
            )
        )
        bucket.blob.assert_called_once_with("my_prefix/full/filename")
        assert blob.cache_control == GCSFilesStore.CACHE_CONTROL
        assert blob.metadata == {"foo": "1"}
        blob.upload_from_string.assert_called_once_with(
            data=b"hello",
            content_type="image/png",
            predefined_acl=store.POLICY,
        )

    @coroutine_test
    async def test_persist_defaults(self) -> None:
        store, _, blob = self.build_gcs_files_store()
        await maybe_deferred_to_future(
            store.persist_file(
                "full/filename",
                BytesIO(b"hello"),
                info=DUMMY_SPIDER_INFO,
            )
        )
        blob.upload_from_string.assert_called_once_with(
            data=b"hello",
            content_type="application/octet-stream",
            predefined_acl=store.POLICY,
        )
        assert blob.metadata == {}

    @coroutine_test
    async def test_stat(self) -> None:
        store, bucket, blob = self.build_gcs_files_store()
        checksum = "cdcda85605e46d0af6110752770dce3c"
        blob.md5_hash = base64.b64encode(bytes.fromhex(checksum)).decode()
        updated = datetime(2019, 12, 1)
        blob.updated = updated
        bucket.get_blob.return_value = blob
        stat = await maybe_deferred_to_future(
            store.stat_file("full/filename", info=DUMMY_SPIDER_INFO)
        )
        bucket.get_blob.assert_called_once_with("my_prefix/full/filename")
        assert stat == {
            "checksum": checksum,
            "last_modified": time.mktime(updated.timetuple()),
        }

    @coroutine_test
    async def test_stat_missing_blob(self) -> None:
        store, bucket, _ = self.build_gcs_files_store()
        bucket.get_blob.return_value = None
        stat = await maybe_deferred_to_future(
            store.stat_file("full/filename", info=DUMMY_SPIDER_INFO)
        )
        assert stat == {}

    @coroutine_test
    async def test_blob_path_consistency(self) -> None:
        """Test to make sure that paths used to store files is the same as the one used to get
        already uploaded files.
        """
        store, bucket, _ = self.build_gcs_files_store()
        bucket.get_blob.return_value = None
        path = "full/my_data.txt"
        await maybe_deferred_to_future(
            store.persist_file(path, BytesIO(b""), info=DUMMY_SPIDER_INFO)
        )
        await maybe_deferred_to_future(store.stat_file(path, info=DUMMY_SPIDER_INFO))
        expected_blob_path = store.prefix + path
        bucket.blob.assert_called_with(expected_blob_path)
        bucket.get_blob.assert_called_with(expected_blob_path)


class TestFTPFileStore:
    @inline_callbacks_test
    def test_persist(self):
        data = b"TestFTPFilesStore: \xe2\x98\x83"
        buf = BytesIO(data)
        meta = {"foo": "bar"}
        path = "full/filename"
        with MockFTPServer() as ftp_server:
            # normally set via FilesPipeline.from_crawler()
            FTPFilesStore.FTP_USERNAME = "anonymous"
            FTPFilesStore.FTP_PASSWORD = "guest"

            store = FTPFilesStore(ftp_server.url("/"))
            empty_dict = yield store.stat_file(path, info=DUMMY_SPIDER_INFO)
            assert empty_dict == {}
            yield store.persist_file(
                path, buf, info=DUMMY_SPIDER_INFO, meta=meta, headers=None
            )
            stat = yield store.stat_file(path, info=DUMMY_SPIDER_INFO)
            assert "last_modified" in stat
            assert "checksum" in stat
            assert stat["checksum"] == "d113d66b2ec7258724a268bd88eef6b6"
            path = f"{store.basedir}/{path}"
            content = get_ftp_content_and_delete(
                path,
                store.host,
                store.port,
                store.username,
                store.password,
                bool(store.USE_ACTIVE_MODE),
            )
        assert data == content

    @inline_callbacks_test
    def test_persist_active_mode(self, monkeypatch: pytest.MonkeyPatch):
        data = b"active mode"
        path = "full/filename"
        monkeypatch.setattr(FTPFilesStore, "FTP_USERNAME", "anonymous")
        monkeypatch.setattr(FTPFilesStore, "FTP_PASSWORD", "guest")
        monkeypatch.setattr(FTPFilesStore, "USE_ACTIVE_MODE", True)
        with MockFTPServer() as ftp_server:
            store = FTPFilesStore(ftp_server.url("/"))
            yield store.persist_file(path, BytesIO(data), info=DUMMY_SPIDER_INFO)
            stat = yield store.stat_file(path, info=DUMMY_SPIDER_INFO)
        assert stat["checksum"] == "ff1575649a39a27c13faa0d37c84bab3"

    def test_wrong_uri_scheme(self):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Incorrect URI scheme in http://example.com/, expected 'ftp'"
            ),
        ):
            FTPFilesStore("http://example.com/")


class ItemWithFiles(Item):
    file_urls = Field()
    files = Field()


def _create_item_with_files(*files: str) -> ItemWithFiles:
    item = ItemWithFiles()
    item["file_urls"] = files
    return item


def _prepare_request_object(
    item_url: str,
    flags: list[str] | None = None,
    headers: dict[str, str] | None = None,
) -> Request:
    return Request(
        item_url,
        meta={
            "response": Response(
                item_url, status=200, body=b"data", flags=flags, headers=headers
            )
        },
    )


# this is separate from the one in test_pipeline_media.py to specifically test FilesPipeline subclasses
class TestBuildFromCrawler:
    def setup_method(self):
        self.tempdir = mkdtemp()
        self.crawler = get_crawler(None, {"FILES_STORE": self.tempdir})

    def teardown_method(self):
        rmtree(self.tempdir)

    def test_simple(self):
        class Pipeline(FilesPipeline):
            pass

        pipe = build_from_crawler(Pipeline, self.crawler)
        assert pipe.crawler == self.crawler
        assert pipe._fingerprinter
        assert pipe.store

    def test_has_from_crawler_and_init(self):
        class Pipeline(FilesPipeline):
            _from_crawler_called = False

            @classmethod
            def from_crawler(cls, crawler: Crawler) -> "Pipeline":
                settings = crawler.settings
                store_uri = settings["FILES_STORE"]
                o = cls(store_uri, crawler=crawler)
                o._from_crawler_called = True
                return o

        pipe = build_from_crawler(Pipeline, self.crawler)
        assert pipe.crawler == self.crawler
        assert pipe._fingerprinter
        assert pipe.store
        assert pipe._from_crawler_called


@pytest.mark.parametrize("store", [None, ""])
def test_files_pipeline_raises_notconfigured_when_files_store_invalid(store):
    settings = Settings()
    settings.clear()
    settings.set("FILES_STORE", store, priority="cmdline")
    crawler = get_crawler(settings_dict=dict(settings))

    with pytest.raises(NotConfigured):
        build_from_crawler(FilesPipeline, crawler)


def test_file_exception_deprecated_import():
    with pytest.warns(ScrapyDeprecationWarning, match="FileException"):
        assert files.FileException is FileException

    with pytest.raises(AttributeError):
        files.nonexistent



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from scrapy.http import Request, Response
from scrapy.spidermiddlewares.depth import DepthMiddleware
from scrapy.spiders import Spider
from scrapy.utils.asyncgen import as_async_generator, collect_asyncgen
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from collections.abc import Generator

    from scrapy.crawler import Crawler
    from scrapy.statscollectors import StatsCollector


@pytest.fixture
def crawler() -> Crawler:
    return get_crawler(Spider, {"DEPTH_LIMIT": 1, "DEPTH_STATS_VERBOSE": True})


@pytest.fixture
def stats(crawler: Crawler) -> Generator[StatsCollector]:
    crawler.stats.open_spider()

    yield crawler.stats

    crawler.stats.close_spider()


@pytest.fixture
def mw(crawler: Crawler) -> DepthMiddleware:
    return build_from_crawler(DepthMiddleware, crawler)


def test_process_spider_output(mw: DepthMiddleware, stats: StatsCollector) -> None:
    req = Request("http://scrapytest.org")
    resp = Response("http://scrapytest.org")
    resp.request = req
    result = [Request("http://scrapytest.org")]

    out = list(mw.process_spider_output(resp, result))
    assert out == result

    rdc = stats.get_value("request_depth_count/1")
    assert rdc == 1

    req.meta["depth"] = 1

    out2 = list(mw.process_spider_output(resp, result))
    assert not out2

    rdm = stats.get_value("request_depth_max")
    assert rdm == 1


def test_depth_reset(mw: DepthMiddleware, stats: StatsCollector) -> None:
    resp = Response("https://example.com")
    resp.request = Request("https://example.com", meta={"depth": 5})
    result = [Request("https://example.com", meta={"depth_reset": True})]

    out = list(mw.process_spider_output(resp, result))

    assert out == result
    assert out[0].meta["depth"] == 0
    assert "depth_reset" not in out[0].meta
    assert stats.get_value("request_depth_count/0") == 1


def test_process_spider_output_no_response(
    mw: DepthMiddleware, stats: StatsCollector
) -> None:
    result = [Request("https://example.com")]

    out = list(mw.process_spider_output(None, result))
    assert out == result
    assert "depth" not in out[0].meta
    assert stats.get_value("request_depth_count/0") is None


@coroutine_test
async def test_process_spider_output_async_no_response(
    mw: DepthMiddleware, stats: StatsCollector
) -> None:
    result = [Request("https://example.com")]

    out = await collect_asyncgen(
        mw.process_spider_output_async(None, as_async_generator(result))
    )
    assert out == result
    assert "depth" not in out[0].meta
    assert stats.get_value("request_depth_count/0") is None


def test_ignored_logged_once(
    mw: DepthMiddleware, stats: StatsCollector, caplog: pytest.LogCaptureFixture
) -> None:
    resp = Response("http://example.com")
    resp.request = Request("http://example.com")
    resp.meta["depth"] = 1
    result = [Request(f"http://example.com/{i}") for i in range(3)]

    with caplog.at_level("DEBUG", logger="scrapy.spidermiddlewares.depth"):
        assert not list(mw.process_spider_output(resp, result))

    messages = [r.getMessage() for r in caplog.records]
    assert len(messages) == 1
    assert "http://example.com/0" in messages[0]
    assert stats.get_value("depth/request_ignored_count") == 3


def test_priority_and_non_verbose_stats() -> None:
    crawler = get_crawler(
        Spider,
        {"DEPTH_LIMIT": 0, "DEPTH_STATS_VERBOSE": False, "DEPTH_PRIORITY": 10},
    )
    crawler.stats.open_spider()
    try:
        mw = build_from_crawler(DepthMiddleware, crawler)
        resp = Response("http://toscrape.com")
        resp.request = Request("http://toscrape.com")
        resp.request.meta["depth"] = 2
        out = list(mw.process_spider_output(resp, [Request("http://toscrape.com")]))
        assert len(out) == 1
        # priority is decremented by depth * DEPTH_PRIORITY
        assert out[0].priority == -30
        assert out[0].meta["depth"] == 3
        # non-verbose stats don't track per-depth counts but still track the max
        assert crawler.stats.get_value("request_depth_count/3") is None
        assert crawler.stats.get_value("request_depth_max") == 3
    finally:
        crawler.stats.close_spider()



"""Download handlers for http and https schemes"""

from __future__ import annotations

import ipaddress
import logging
import re
from contextlib import suppress
from functools import partial
from io import BytesIO
from time import monotonic
from typing import TYPE_CHECKING, Any, TypedDict, TypeVar, cast
from urllib.parse import urldefrag, urlparse

from twisted.internet import ssl
from twisted.internet.defer import Deferred, succeed
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.internet.protocol import Factory, Protocol, connectionDone
from twisted.python.failure import Failure
from twisted.web._newclient import (
    HEADER,
    STATUS,
    HTTP11ClientProtocol,
    HTTPClientParser,
)
from twisted.web.client import (
    URI,
    Agent,
    HTTPConnectionPool,
    ResponseDone,
    ResponseFailed,
    _HTTP11ClientFactory,
)
from twisted.web.client import Response as TxResponse
from twisted.web.http import PotentialDataLoss, _DataLoss
from twisted.web.http_headers import Headers as TxHeaders
from twisted.web.iweb import UNKNOWN_LENGTH, IBodyProducer, IPolicyForHTTPS, IResponse
from zope.interface import implementer

from scrapy import Request, signals
from scrapy.core.downloader._idna_patch import _install_twisted_idna_fallbacks
from scrapy.core.downloader.contextfactory import _load_context_factory_from_settings
from scrapy.exceptions import (
    DownloadCancelledError,
    DownloadTimeoutError,
    NotConfigured,
    ResponseDataLossError,
    StopDownload,
)
from scrapy.http import Headers, Response
from scrapy.utils._download_handlers import (
    check_stop_download,
    get_dataloss_msg,
    get_maxsize_msg,
    get_warnsize_msg,
    make_response,
    normalize_bind_address,
    wrap_twisted_exceptions,
)
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.deprecate import warn_on_deprecated_spider_attribute
from scrapy.utils.httpobj import urlparse_cached
from scrapy.utils.python import to_bytes, to_unicode
from scrapy.utils.ssl import _log_ssl_conn_debug_info
from scrapy.utils.url import add_http_if_no_scheme

from ._base_http import BaseHttpDownloadHandler

if TYPE_CHECKING:
    from twisted.internet.base import ReactorBase
    from twisted.internet.interfaces import IAddress, IConsumer
    from twisted.web._newclient import Request as TxRequest

    # typing.NotRequired requires Python 3.11
    from typing_extensions import NotRequired

    from scrapy.crawler import Crawler


logger = logging.getLogger(__name__)

_T = TypeVar("_T")


class _ResultT(TypedDict):
    txresponse: TxResponse
    body: NotRequired[bytes]
    flags: NotRequired[list[str] | None]
    certificate: NotRequired[ssl.Certificate | None]
    ip_address: NotRequired[ipaddress.IPv4Address | ipaddress.IPv6Address | None]
    stop_download: NotRequired[StopDownload | None]


class HTTP11DownloadHandler(BaseHttpDownloadHandler):
    def __init__(self, crawler: Crawler):
        if not crawler.settings.getbool("TWISTED_REACTOR_ENABLED"):
            raise NotConfigured(f"{type(self).__name__} requires a Twisted reactor.")
        _install_twisted_idna_fallbacks()
        super().__init__(crawler)
        self._crawler = crawler

        from twisted.internet import reactor

        self._pool: HTTPConnectionPool = HTTPConnectionPool(reactor, persistent=True)
        self._pool.maxPersistentPerHost = crawler.settings.getint(
            "CONCURRENT_REQUESTS_PER_DOMAIN"
        )
        self._pool._factory = _LenientHTTP11ClientFactory

        self._contextFactory: IPolicyForHTTPS = _load_context_factory_from_settings(
            crawler
        )
        self._bind_address = crawler.settings.get("DOWNLOAD_BIND_ADDRESS")
        self._disconnect_timeout: int = 1

    async def download_request(self, request: Request) -> Response:
        if hasattr(self._crawler.spider, "download_maxsize"):  # pragma: no cover
            warn_on_deprecated_spider_attribute("download_maxsize", "DOWNLOAD_MAXSIZE")
        if hasattr(self._crawler.spider, "download_warnsize"):  # pragma: no cover
            warn_on_deprecated_spider_attribute(
                "download_warnsize", "DOWNLOAD_WARNSIZE"
            )

        agent = _ScrapyAgent(
            contextFactory=self._contextFactory,
            bindAddress=self._bind_address,
            pool=self._pool,
            maxsize=getattr(
                self._crawler.spider, "download_maxsize", self._default_maxsize
            ),
            warnsize=getattr(
                self._crawler.spider, "download_warnsize", self._default_warnsize
            ),
            fail_on_dataloss=self._fail_on_dataloss,
            crawler=self._crawler,
            tls_verbose_logging=self._tls_verbose_logging,
        )
        try:
            with wrap_twisted_exceptions():
                return await maybe_deferred_to_future(agent.download_request(request))
        except ResponseDataLossError:
            if not self._fail_on_dataloss_warned:
                logger.warning(get_dataloss_msg(request.url))
                self._fail_on_dataloss_warned = True
            raise

    async def close(self) -> None:
        from twisted.internet import reactor

        d: Deferred[None] = self._pool.closeCachedConnections()
        # closeCachedConnections will hang on network or server issues, so
        # we'll manually timeout the deferred.
        #
        # Twisted issue addressing this problem can be found here:
        # https://github.com/twisted/twisted/issues/7738
        #
        # closeCachedConnections doesn't handle external errbacks, so we'll
        # issue a callback after `_disconnect_timeout` seconds.
        #
        # See also https://github.com/scrapy/scrapy/issues/2653
        delayed_call = reactor.callLater(self._disconnect_timeout, d.callback, ())

        try:
            await maybe_deferred_to_future(d)
        finally:
            if delayed_call.active():
                delayed_call.cancel()


class TunnelError(Exception):
    """An HTTP CONNECT tunnel could not be established by the proxy."""


class _TunnelingTCP4ClientEndpoint(TCP4ClientEndpoint):
    """An endpoint that tunnels through proxies to allow HTTPS downloads. To
    accomplish that, this endpoint sends an HTTP CONNECT to the proxy.
    The HTTP CONNECT is always sent when using this endpoint, I think this could
    be improved as the CONNECT will be redundant if the connection associated
    with this endpoint comes from the pool and a CONNECT has already been issued
    for it.
    """

    _truncatedLength = 1000
    _responseAnswer = (
        r"HTTP/1\.. (?P<status>\d{3})(?P<reason>.{," + str(_truncatedLength) + r"})"
    )
    _responseMatcher = re.compile(_responseAnswer.encode())

    def __init__(
        self,
        reactor: ReactorBase,
        host: str,
        port: int,
        proxyConf: tuple[str, int, bytes | None],
        contextFactory: IPolicyForHTTPS,
        timeout: float = 30,
        bindAddress: tuple[str, int] | None = None,
    ):
        proxyHost, proxyPort, self._proxyAuthHeader = proxyConf
        super().__init__(reactor, proxyHost, proxyPort, timeout, bindAddress)
        self._tunnelReadyDeferred: Deferred[Protocol] = Deferred()
        self._tunneledHost: str = host
        self._tunneledPort: int = port
        self._contextFactory: IPolicyForHTTPS = contextFactory
        self._connectBuffer: bytearray = bytearray()

    def requestTunnel(self, protocol: Protocol) -> Protocol:
        """Asks the proxy to open a tunnel."""
        assert protocol.transport
        tunnelReq = _tunnel_request_data(
            self._tunneledHost, self._tunneledPort, self._proxyAuthHeader
        )
        protocol.transport.write(tunnelReq)
        self._protocolDataReceived = protocol.dataReceived
        protocol.dataReceived = self.processProxyResponse  # type: ignore[method-assign]
        self._protocol = protocol
        return protocol

    def processProxyResponse(self, data: bytes) -> None:
        """Processes the response from the proxy. If the tunnel is successfully
        created, notifies the client that we are ready to send requests. If not
        raises a TunnelError.
        """
        assert self._protocol.transport
        self._connectBuffer += data
        # make sure that enough (all) bytes are consumed
        # and that we've got all HTTP headers (ending with a blank line)
        # from the proxy so that we don't send those bytes to the TLS layer
        #
        # see https://github.com/scrapy/scrapy/issues/2491
        if b"\r\n\r\n" not in self._connectBuffer:
            return
        self._protocol.dataReceived = self._protocolDataReceived  # type: ignore[method-assign]
        respm = _TunnelingTCP4ClientEndpoint._responseMatcher.match(self._connectBuffer)
        if respm and int(respm.group("status")) == 200:
            # set proper Server Name Indication extension
            sslOptions = self._contextFactory.creatorForNetloc(  # type: ignore[call-arg,misc]
                self._tunneledHost,  # type: ignore[arg-type]
                self._tunneledPort,
            )
            self._protocol.transport.startTLS(sslOptions, self._protocolFactory)
            self._tunnelReadyDeferred.callback(self._protocol)
        else:
            extra: Any
            if respm:
                extra = {
                    "status": int(respm.group("status")),
                    "reason": respm.group("reason").strip(),
                }
            else:
                extra = data[: self._truncatedLength]
            self._tunnelReadyDeferred.errback(
                TunnelError(
                    "Could not open CONNECT tunnel with proxy "
                    f"{self._host}:{self._port} [{extra!r}]"
                )
            )

    def connectFailed(self, reason: Failure) -> None:
        """Propagates the errback to the appropriate deferred."""
        self._tunnelReadyDeferred.errback(reason)

    def connect(self, protocolFactory: Factory) -> Deferred[Protocol]:
        self._protocolFactory = protocolFactory
        connectDeferred = super().connect(protocolFactory)
        connectDeferred.addCallback(self.requestTunnel)
        connectDeferred.addErrback(self.connectFailed)
        return self._tunnelReadyDeferred


def _tunnel_request_data(
    host: str, port: int, proxy_auth_header: bytes | None = None
) -> bytes:
    r"""
    Return binary content of a CONNECT request.

    >>> from scrapy.utils.python import to_unicode as s
    >>> s(_tunnel_request_data("example.com", 8080))
    'CONNECT example.com:8080 HTTP/1.1\r\nHost: example.com:8080\r\n\r\n'
    >>> s(_tunnel_request_data("example.com", 8080, b"123"))
    'CONNECT example.com:8080 HTTP/1.1\r\nHost: example.com:8080\r\nProxy-Authorization: 123\r\n\r\n'
    >>> s(_tunnel_request_data(b"example.com", "8090"))
    'CONNECT example.com:8090 HTTP/1.1\r\nHost: example.com:8090\r\n\r\n'
    """
    host_value = to_bytes(host, encoding="ascii") + b":" + to_bytes(str(port))
    tunnel_req = b"CONNECT " + host_value + b" HTTP/1.1\r\n"
    tunnel_req += b"Host: " + host_value + b"\r\n"
    if proxy_auth_header:
        tunnel_req += b"Proxy-Authorization: " + proxy_auth_header + b"\r\n"
    tunnel_req += b"\r\n"
    return tunnel_req


class _TunnelingAgent(Agent):
    """An agent that uses a ``_TunnelingTCP4ClientEndpoint`` to make HTTPS
    downloads. It may look strange that we have chosen to subclass Agent and not
    ProxyAgent but consider that after the tunnel is opened the proxy is
    transparent to the client; thus the agent should behave like there is no
    proxy involved.
    """

    def __init__(
        self,
        *,
        reactor: ReactorBase,
        proxyConf: tuple[str, int, bytes | None],
        contextFactory: IPolicyForHTTPS,
        connectTimeout: float | None = None,
        bindAddress: tuple[str, int] | None = None,
        pool: HTTPConnectionPool | None = None,
    ):
        super().__init__(reactor, contextFactory, connectTimeout, bindAddress, pool)  # type: ignore[no-untyped-call]
        self._proxyConf: tuple[str, int, bytes | None] = proxyConf
        self._contextFactory: IPolicyForHTTPS = contextFactory

    def _getEndpoint(self, uri: URI) -> _TunnelingTCP4ClientEndpoint:
        return _TunnelingTCP4ClientEndpoint(
            reactor=self._reactor,
            host=uri.host,
            port=uri.port,
            proxyConf=self._proxyConf,
            contextFactory=self._contextFactory,
            timeout=self._endpointFactory._connectTimeout,
            bindAddress=self._endpointFactory._bindAddress,
        )

    def _requestWithEndpoint(
        self,
        key: Any,
        endpoint: TCP4ClientEndpoint,
        method: bytes,
        parsedURI: URI,
        headers: TxHeaders | None,
        bodyProducer: IBodyProducer | None,
        requestPath: bytes,
    ) -> Deferred[IResponse]:
        # proxy host and port are required for HTTP pool `key`
        # otherwise, same remote host connection request could reuse
        # a cached tunneled connection to a different proxy
        key += self._proxyConf
        return super()._requestWithEndpoint(
            key=key,
            endpoint=endpoint,
            method=method,
            parsedURI=parsedURI,
            headers=headers,
            bodyProducer=bodyProducer,
            requestPath=requestPath,
        )


class _ScrapyProxyAgent(Agent):
    def __init__(
        self,
        reactor: ReactorBase,
        proxyURI: bytes,
        contextFactory: IPolicyForHTTPS,
        connectTimeout: float | None = None,
        bindAddress: tuple[str, int] | None = None,
        pool: HTTPConnectionPool | None = None,
    ):
        super().__init__(  # type: ignore[no-untyped-call]
            reactor=reactor,
            contextFactory=contextFactory,
            connectTimeout=connectTimeout,
            bindAddress=bindAddress,
            pool=pool,
        )
        self._proxyURI: URI = URI.fromBytes(proxyURI)

    def request(
        self,
        method: bytes,
        uri: bytes,
        headers: TxHeaders | None = None,
        bodyProducer: IBodyProducer | None = None,
    ) -> Deferred[IResponse]:
        """
        Issue a new request via the configured proxy.
        """
        # Cache *all* connections under the same key, since we are only
        # connecting to a single destination, the proxy:
        return self._requestWithEndpoint(
            key=(b"http-proxy", self._proxyURI.host, self._proxyURI.port),
            endpoint=self._getEndpoint(self._proxyURI),  # type: ignore[no-untyped-call]
            method=method,
            parsedURI=URI.fromBytes(uri),
            headers=headers,
            bodyProducer=bodyProducer,
            requestPath=uri,
        )


class _ScrapyAgent:
    def __init__(
        self,
        *,
        contextFactory: IPolicyForHTTPS,
        connectTimeout: float = 10,
        bindAddress: str | tuple[str, int] | None = None,
        pool: HTTPConnectionPool | None = None,
        maxsize: int = 0,
        warnsize: int = 0,
        fail_on_dataloss: bool = True,
        crawler: Crawler,
        tls_verbose_logging: bool = False,
    ):
        self._contextFactory: IPolicyForHTTPS = contextFactory
        self._connectTimeout: float = connectTimeout
        self._bindAddress: str | tuple[str, int] | None = bindAddress
        self._pool: HTTPConnectionPool | None = pool
        self._maxsize: int = maxsize
        self._warnsize: int = warnsize
        self._fail_on_dataloss: bool = fail_on_dataloss
        self._txresponse: TxResponse | None = None
        self._crawler: Crawler = crawler
        self._tls_verbose_logging: bool = tls_verbose_logging

    def _get_agent(self, request: Request, timeout: float) -> Agent:
        from twisted.internet import reactor

        bindaddress = request.meta.get("bindaddress") or self._bindAddress
        bindaddress = normalize_bind_address(bindaddress)
        proxy = request.meta.get("proxy")
        if proxy:
            proxy = add_http_if_no_scheme(proxy)
            proxy_parsed = urlparse(proxy)
            proxy_host = proxy_parsed.hostname
            proxy_port = proxy_parsed.port
            if not proxy_port:
                proxy_port = 443 if proxy_parsed.scheme == "https" else 80
            if urlparse_cached(request).scheme == "https":
                if proxy_parsed.scheme == "https":  # pragma: no cover
                    raise NotImplementedError(
                        "HTTPS proxies for HTTPS destinations are not supported"
                    )
                assert proxy_host is not None
                proxyAuth = request.headers.get(b"Proxy-Authorization", None)
                proxyConf = (proxy_host, proxy_port, proxyAuth)
                return _TunnelingAgent(
                    reactor=reactor,
                    proxyConf=proxyConf,
                    contextFactory=self._contextFactory,
                    connectTimeout=timeout,
                    bindAddress=bindaddress,
                    pool=self._pool,
                )
            return _ScrapyProxyAgent(
                reactor=reactor,
                proxyURI=to_bytes(proxy, encoding="ascii"),
                contextFactory=self._contextFactory,
                connectTimeout=timeout,
                bindAddress=bindaddress,
                pool=self._pool,
            )

        return Agent(
            reactor=reactor,
            contextFactory=self._contextFactory,
            connectTimeout=timeout,
            bindAddress=bindaddress,
            pool=self._pool,
        )

    def download_request(self, request: Request) -> Deferred[Response]:
        from twisted.internet import reactor

        timeout = request.meta.get("download_timeout") or self._connectTimeout
        agent = self._get_agent(request, timeout)

        # request details
        url = urldefrag(request.url)[0]
        method = to_bytes(request.method)
        headers = TxHeaders(request.headers)
        if isinstance(agent, _TunnelingAgent):
            headers.removeHeader(b"Proxy-Authorization")
        bodyproducer = _RequestBodyProducer(request.body) if request.body else None
        start_time = monotonic()
        d: Deferred[IResponse] = agent.request(
            method,
            to_bytes(url, encoding="ascii"),
            headers,
            cast("IBodyProducer", bodyproducer),
        )
        # set download latency
        d.addCallback(self._cb_latency, request, start_time)
        # response body is ready to be consumed
        d2: Deferred[_ResultT] = d.addCallback(self._cb_bodyready, request)
        d3: Deferred[Response] = d2.addCallback(self._cb_bodydone, url)
        # check download timeout
        self._timeout_cl = reactor.callLater(timeout, d3.cancel)
        d3.addBoth(self._cb_timeout, request, url, timeout)
        return d3

    def _cb_timeout(self, result: _T, request: Request, url: str, timeout: float) -> _T:
        if self._timeout_cl.active():
            self._timeout_cl.cancel()
            return result
        # needed for HTTPS requests, otherwise _ResponseReader doesn't
        # receive connectionLost()
        if self._txresponse:
            self._txresponse._transport.stopProducing()

        raise DownloadTimeoutError(f"Getting {url} took longer than {timeout} seconds.")

    def _cb_latency(self, result: _T, request: Request, start_time: float) -> _T:
        request.meta["download_latency"] = monotonic() - start_time
        return result

    @staticmethod
    def _headers_from_twisted_response(response: TxResponse) -> Headers:
        headers = Headers()
        if response.length != UNKNOWN_LENGTH:
            headers[b"Content-Length"] = str(response.length).encode()
        headers.update(response.headers.getAllRawHeaders())
        return headers

    def _cb_bodyready(
        self, txresponse: TxResponse, request: Request
    ) -> _ResultT | Deferred[_ResultT]:
        if stop_download := check_stop_download(
            signals.headers_received,
            self._crawler,
            request,
            headers=self._headers_from_twisted_response(txresponse),
            body_length=txresponse.length,
        ):
            txresponse._transport.stopProducing()
            txresponse._transport.loseConnection()
            return {
                "txresponse": txresponse,
                "stop_download": stop_download,
            }

        # deliverBody hangs for responses without body
        if cast("int", txresponse.length) == 0:
            return {
                "txresponse": txresponse,
            }

        maxsize = request.meta.get("download_maxsize", self._maxsize)
        warnsize = request.meta.get("download_warnsize", self._warnsize)
        expected_size = (
            cast("int", txresponse.length)
            if txresponse.length != UNKNOWN_LENGTH
            else -1
        )
        fail_on_dataloss = request.meta.get(
            "download_fail_on_dataloss", self._fail_on_dataloss
        )

        if maxsize and expected_size > maxsize:
            warning_msg = get_maxsize_msg(
                expected_size, maxsize, request, expected=True
            )
            logger.warning(warning_msg)
            # Abort connection immediately.
            txresponse._transport._producer.abortConnection()
            raise DownloadCancelledError(warning_msg)

        reached_warnsize = bool(warnsize and expected_size > warnsize)
        if reached_warnsize:
            logger.warning(
                get_warnsize_msg(expected_size, warnsize, request, expected=True)
            )

        d: Deferred[_ResultT] = Deferred(partial(self._cancel, txresponse=txresponse))
        txresponse.deliverBody(
            _ResponseReader(
                finished=d,
                txresponse=txresponse,
                request=request,
                maxsize=maxsize,
                warnsize=warnsize,
                reached_warnsize=reached_warnsize,
                fail_on_dataloss=fail_on_dataloss,
                crawler=self._crawler,
                tls_verbose_logging=self._tls_verbose_logging,
            )
        )

        # save response for timeouts
        self._txresponse = txresponse

        return d

    @staticmethod
    def _cancel(_: Any, txresponse: TxResponse) -> None:
        # Abort connection immediately.
        txresponse._transport._producer.abortConnection()

    def _cb_bodydone(self, result: _ResultT, url: str) -> Response:
        headers = self._headers_from_twisted_response(result["txresponse"])
        try:
            version = result["txresponse"].version
            protocol = f"{to_unicode(version[0])}/{version[1]}.{version[2]}"
        except (AttributeError, TypeError, IndexError):
            protocol = None
        return make_response(
            url=url,
            status=int(result["txresponse"].code),
            headers=headers,
            body=result.get("body", b""),
            flags=result.get("flags"),
            certificate=result.get("certificate"),
            ip_address=result.get("ip_address"),
            protocol=protocol,
            stop_download=result.get("stop_download"),
        )


@implementer(IBodyProducer)
class _RequestBodyProducer:
    def __init__(self, body: bytes):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer: IConsumer) -> Deferred[None]:
        consumer.write(self.body)
        return succeed(None)

    def pauseProducing(self) -> None:
        pass

    def stopProducing(self) -> None:
        pass


class _ResponseReader(Protocol):
    def __init__(
        self,
        finished: Deferred[_ResultT],
        txresponse: TxResponse,
        request: Request,
        maxsize: int,
        warnsize: int,
        fail_on_dataloss: bool,
        crawler: Crawler,
        *,
        reached_warnsize: bool = False,
        tls_verbose_logging: bool = False,
    ):
        self._finished: Deferred[_ResultT] = finished
        self._txresponse: TxResponse = txresponse
        self._request: Request = request
        self._bodybuf: BytesIO = BytesIO()
        self._maxsize: int = maxsize
        self._warnsize: int = warnsize
        self._fail_on_dataloss: bool = fail_on_dataloss
        self._reached_warnsize: bool = reached_warnsize
        self._bytes_received: int = 0
        self._certificate: ssl.Certificate | None = None
        self._ip_address: ipaddress.IPv4Address | ipaddress.IPv6Address | None = None
        self._crawler: Crawler = crawler
        self._tls_verbose_logging: bool = tls_verbose_logging

    def _finish_response(
        self, flags: list[str] | None = None, stop_download: StopDownload | None = None
    ) -> None:
        self._finished.callback(
            {
                "txresponse": self._txresponse,
                "body": self._bodybuf.getvalue(),
                "flags": flags,
                "certificate": self._certificate,
                "ip_address": self._ip_address,
                "stop_download": stop_download,
            }
        )

    def connectionMade(self) -> None:
        assert self.transport
        if self._certificate is None:
            with suppress(AttributeError):
                self._certificate = ssl.Certificate(
                    self.transport._producer.getPeerCertificate()
                )

        if self._ip_address is None:
            self._ip_address = ipaddress.ip_address(
                self.transport._producer.getPeer().host
            )

        if self._tls_verbose_logging:
            connection = self.transport._producer.getHandle()
            hostname = urlparse_cached(self._request).hostname
            assert hostname is not None
            _log_ssl_conn_debug_info(hostname, connection)

    def dataReceived(self, data: bytes) -> None:
        # This maybe called several times after cancel was called with buffered data.
        if self._finished.called:
            return

        assert self.transport
        self._bodybuf.write(data)
        self._bytes_received += len(data)

        if stop_download := check_stop_download(
            signals.bytes_received, self._crawler, self._request, data=data
        ):
            self.transport.stopProducing()
            self.transport.loseConnection()
            self._finish_response(stop_download=stop_download)

        if self._maxsize and self._bytes_received > self._maxsize:
            logger.warning(
                get_maxsize_msg(
                    self._bytes_received, self._maxsize, self._request, expected=False
                )
            )
            # Clear buffer earlier to avoid keeping data in memory for a long time.
            self._bodybuf.truncate(0)
            self._finished.cancel()

        if (
            self._warnsize
            and self._bytes_received > self._warnsize
            and not self._reached_warnsize
        ):
            self._reached_warnsize = True
            logger.warning(
                get_warnsize_msg(
                    self._bytes_received, self._warnsize, self._request, expected=False
                )
            )

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        if self._finished.called:
            return

        if reason.check(ResponseDone):
            self._finish_response()
            return

        if reason.check(PotentialDataLoss):
            self._finish_response(flags=["partial"])
            return

        if reason.check(ResponseFailed) and any(
            r.check(_DataLoss)
            for r in reason.value.reasons  # type: ignore[union-attr]
        ):
            if not self._fail_on_dataloss:
                self._finish_response(flags=["dataloss"])
                return

            exc = ResponseDataLossError()
            exc.__cause__ = reason.value
            reason = Failure(exc)

        self._finished.errback(reason)


class _LenientHTTPClientParser(HTTPClientParser):
    """Response parser that skips bad response header lines, those with no
    colon in them, instead of failing to parse the whole response.

    Some servers send such lines, and web browsers skip them and keep parsing
    the header lines that follow. See
    https://github.com/scrapy/scrapy/issues/210.
    """

    def lineReceived(self, line: bytes) -> None:
        # A copy of twisted.web._newclient.HTTPParser.lineReceived() where the
        # header name and value are only extracted from header lines that have
        # a colon.

        # Handle the normal CR LF case.
        if line[-1:] == b"\r":
            line = line[:-1]

        if self.state == STATUS:
            self.statusReceived(line)  # type: ignore[no-untyped-call]
            self.state = HEADER
            return

        # HEADER is the only other state in which lines are received, as the
        # parser switches to raw mode for the response body.
        if not line or line[0] not in b" \t":
            if self._partialHeader is not None:
                header = b"".join(self._partialHeader)
                if b":" in header:
                    name, value = header.split(b":", 1)
                    self.headerReceived(name, value.strip())  # type: ignore[no-untyped-call]
                else:
                    logger.debug(
                        f"Skipping the bad response header line {header!r}, as "
                        f"it has no colon."
                    )
            if not line:
                # Empty line means the header section is over.
                self.allHeadersReceived()  # type: ignore[no-untyped-call]
            else:
                # Line not beginning with LWS is another header.
                self._partialHeader = [line]
        else:
            # A line beginning with LWS is a continuation of a header begun on
            # a previous line.
            self._partialHeader.append(line)  # type: ignore[union-attr]


class _LenientHTTP11ClientProtocol(HTTP11ClientProtocol):
    """Protocol that parses responses with :class:`_LenientHTTPClientParser`."""

    def request(self, request: TxRequest) -> Deferred[IResponse]:
        d: Deferred[IResponse] = super().request(request)
        # HTTP11ClientProtocol.request() hardcodes the parser class, so the
        # only way to use a different one is to replace the class of the parser
        # object that it creates. This is safe because
        # _LenientHTTPClientParser defines no additional state. The parser is
        # always there because HTTPConnectionPool only reuses connections whose
        # protocol is in the QUIESCENT state, for which request() always
        # creates a parser.
        assert self._parser is not None
        self._parser.__class__ = _LenientHTTPClientParser
        return d


class _LenientHTTP11ClientFactory(_HTTP11ClientFactory):
    """Factory that builds :class:`_LenientHTTP11ClientProtocol` protocols."""

    noisy = False

    def buildProtocol(self, addr: IAddress | None) -> HTTP11ClientProtocol:
        return _LenientHTTP11ClientProtocol(self._quiescentCallback)  # type: ignore[no-untyped-call]



from __future__ import annotations

import json
import logging
import re
from ipaddress import IPv4Address
from socket import gethostbyname
from typing import TYPE_CHECKING, Any
from urllib.parse import urlencode, urlparse

import pytest
from cryptography.x509 import load_der_x509_certificate
from twisted.internet.defer import succeed
from twisted.internet.ssl import Certificate
from twisted.python.failure import Failure

from scrapy import Spider, signals
from scrapy.exceptions import CloseSpider, ScrapyDeprecationWarning, StopDownload
from scrapy.http import Request
from scrapy.http.response import Response
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.engine import (
    format_engine_status,
    get_engine_status,
    print_engine_status,
)
from scrapy.utils.python import to_unicode
from scrapy.utils.test import get_crawler
from tests import NON_EXISTING_RESOLVABLE
from tests.spiders import (
    AsyncDefAsyncioGenComplexSpider,
    AsyncDefAsyncioGenExcSpider,
    AsyncDefAsyncioGenLoopSpider,
    AsyncDefAsyncioGenSpider,
    AsyncDefAsyncioReqsReturnSpider,
    AsyncDefAsyncioReturnSingleElementSpider,
    AsyncDefAsyncioReturnSpider,
    AsyncDefAsyncioSpider,
    AsyncDefDeferredDirectSpider,
    AsyncDefDeferredMaybeWrappedSpider,
    AsyncDefDeferredWrappedSpider,
    AsyncDefSpider,
    BrokenStartSpider,
    BytesReceivedCallbackSpider,
    BytesReceivedErrbackSpider,
    CrawlSpiderWithAsyncCallback,
    CrawlSpiderWithAsyncGeneratorCallback,
    CrawlSpiderWithErrback,
    CrawlSpiderWithoutErrback,
    CrawlSpiderWithParseMethod,
    CrawlSpiderWithProcessRequestCallbackKeywordArguments,
    DelaySpider,
    DuplicateStartSpider,
    FollowAllSpider,
    HeadersReceivedCallbackSpider,
    HeadersReceivedErrbackSpider,
    SimpleSpider,
    SingleRequestSpider,
    StartGoodAndBadOutput,
    StartItemSpider,
)
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from scrapy.statscollectors import StatsCollector
    from tests.mockserver.http import MockServer


class TestCrawl:
    @coroutine_test
    async def test_follow_all(self, mockserver: MockServer) -> None:
        crawler = get_crawler(FollowAllSpider)
        await crawler.crawl_async(mockserver=mockserver)
        assert isinstance(crawler.spider, FollowAllSpider)
        assert len(crawler.spider.urls_visited) == 11  # 10 + start_url

    @coroutine_test
    async def test_fixed_delay(self, mockserver: MockServer) -> None:
        await self._test_delay(mockserver, total=3, delay=0.2)

    @coroutine_test
    async def test_randomized_delay(self, mockserver: MockServer) -> None:
        await self._test_delay(mockserver, total=3, delay=0.1, randomize=True)

    @staticmethod
    async def _test_delay(
        mockserver: MockServer, total: int, delay: float, randomize: bool = False
    ) -> None:
        crawl_kwargs = {
            "maxlatency": delay * 2,
            "mockserver": mockserver,
            "total": total,
        }
        tolerance = 1 - (0.6 if randomize else 0.2)

        settings = {"DOWNLOAD_DELAY": delay, "RANDOMIZE_DOWNLOAD_DELAY": randomize}
        crawler = get_crawler(FollowAllSpider, settings)
        await crawler.crawl_async(**crawl_kwargs)
        assert crawler.spider
        assert isinstance(crawler.spider, FollowAllSpider)
        times = crawler.spider.times
        total_time = times[-1] - times[0]
        average = total_time / (len(times) - 1)
        assert average > delay * tolerance, f"download delay too small: {average}"

        # Ensure that the same test parameters would cause a failure if no
        # download delay is set. Otherwise, it means we are using a combination
        # of ``total`` and ``delay`` values that are too small for the test
        # code above to have any meaning.
        settings["DOWNLOAD_DELAY"] = 0
        crawler = get_crawler(FollowAllSpider, settings)
        await crawler.crawl_async(**crawl_kwargs)
        assert crawler.spider
        assert isinstance(crawler.spider, FollowAllSpider)
        times = crawler.spider.times
        total_time = times[-1] - times[0]
        average = total_time / (len(times) - 1)
        assert average <= delay / tolerance, "test total or delay values are too small"

    @coroutine_test
    async def test_timeout_success(self, mockserver: MockServer) -> None:
        crawler = get_crawler(DelaySpider)
        await crawler.crawl_async(n=0.5, mockserver=mockserver)
        assert isinstance(crawler.spider, DelaySpider)
        assert crawler.spider.t1 > 0
        assert crawler.spider.t2 > 0
        assert crawler.spider.t2 > crawler.spider.t1

    @coroutine_test
    async def test_timeout_failure(self, mockserver: MockServer) -> None:
        crawler = get_crawler(DelaySpider, {"DOWNLOAD_TIMEOUT": 0.35})
        await crawler.crawl_async(n=0.5, mockserver=mockserver)
        assert isinstance(crawler.spider, DelaySpider)
        assert crawler.spider.t1 > 0
        assert crawler.spider.t2 == 0
        assert crawler.spider.t2_err > 0
        assert crawler.spider.t2_err > crawler.spider.t1

        # server hangs after receiving response headers
        crawler = get_crawler(DelaySpider, {"DOWNLOAD_TIMEOUT": 0.35})
        await crawler.crawl_async(n=0.5, b=1, mockserver=mockserver)
        assert isinstance(crawler.spider, DelaySpider)
        assert crawler.spider.t1 > 0
        assert crawler.spider.t2 == 0
        assert crawler.spider.t2_err > 0
        assert crawler.spider.t2_err > crawler.spider.t1

    @coroutine_test
    async def test_retry_503(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(SimpleSpider)
        with caplog.at_level(logging.DEBUG):
            await crawler.crawl_async(
                mockserver.url("/status?n=503"), mockserver=mockserver
            )
        self._assert_retried(caplog.text)

    @coroutine_test
    async def test_retry_conn_failed(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(SimpleSpider)
        with caplog.at_level(logging.DEBUG):
            await crawler.crawl_async(
                "http://localhost:65432/status?n=503", mockserver=mockserver
            )
        self._assert_retried(caplog.text)

    @coroutine_test
    async def test_retry_dns_error(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        if NON_EXISTING_RESOLVABLE:
            pytest.skip("Non-existing hosts are resolvable")
        crawler = get_crawler(SimpleSpider)
        with caplog.at_level(logging.DEBUG):
            # try to fetch the homepage of a nonexistent domain
            await crawler.crawl_async(
                "http://dns.resolution.invalid./", mockserver=mockserver
            )
        self._assert_retried(caplog.text)

    @coroutine_test
    async def test_start_bug_before_yield(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        with caplog.at_level(logging.ERROR):
            crawler = get_crawler(BrokenStartSpider)
            await crawler.crawl_async(fail_before_yield=1, mockserver=mockserver)

        assert len(caplog.records) == 1
        record = caplog.records[0]
        assert record.exc_info is not None
        assert record.exc_info[0] is ZeroDivisionError

    @coroutine_test
    async def test_start_bug_yielding(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        with caplog.at_level(logging.ERROR):
            crawler = get_crawler(BrokenStartSpider)
            await crawler.crawl_async(fail_yielding=1, mockserver=mockserver)

        assert len(caplog.records) == 1
        record = caplog.records[0]
        assert record.exc_info is not None
        assert record.exc_info[0] is ZeroDivisionError

    @coroutine_test
    async def test_start_items(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        items = []

        def _on_item_scraped(item):
            items.append(item)

        with caplog.at_level(logging.ERROR):
            crawler = get_crawler(StartItemSpider)
            crawler.signals.connect(_on_item_scraped, signals.item_scraped)
            await crawler.crawl_async(mockserver=mockserver)

        assert len(caplog.records) == 0
        assert items == [{"name": "test item"}]

    @coroutine_test
    async def test_start_unsupported_output(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        """Anything that is not a request is assumed to be an item, avoiding a
        potentially expensive call to itemadapter.is_item(), and letting
        instead things fail when ItemAdapter is actually used on the
        corresponding non-item object."""

        items = []

        def _on_item_scraped(item):
            items.append(item)

        with caplog.at_level(logging.ERROR):
            crawler = get_crawler(StartGoodAndBadOutput)
            crawler.signals.connect(_on_item_scraped, signals.item_scraped)
            await crawler.crawl_async(mockserver=mockserver)

        assert len(caplog.records) == 0
        assert len(items) == 3
        assert not any(isinstance(item, Request) for item in items)

    @coroutine_test
    async def test_start_dupes(self, mockserver: MockServer) -> None:
        settings = {"CONCURRENT_REQUESTS": 1}
        crawler = get_crawler(DuplicateStartSpider, settings)
        await crawler.crawl_async(
            dont_filter=True, distinct_urls=2, dupe_factor=3, mockserver=mockserver
        )
        assert isinstance(crawler.spider, DuplicateStartSpider)
        assert crawler.spider.visited == 6

        crawler = get_crawler(DuplicateStartSpider, settings)
        await crawler.crawl_async(
            dont_filter=False,
            distinct_urls=3,
            dupe_factor=4,
            mockserver=mockserver,
        )
        assert isinstance(crawler.spider, DuplicateStartSpider)
        assert crawler.spider.visited == 3

    @coroutine_test
    async def test_unbounded_response(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        # Completeness of responses without Content-Length or Transfer-Encoding
        # can not be determined, we treat them as valid but flagged as "partial"
        query = urlencode(
            {
                "raw": """\
HTTP/1.1 200 OK
Server: Apache-Coyote/1.1
X-Powered-By: Servlet 2.4; JBoss-4.2.3.GA (build: SVNTag=JBoss_4_2_3_GA date=200807181417)/JBossWeb-2.0
Set-Cookie: JSESSIONID=08515F572832D0E659FD2B0D8031D75F; Path=/
Pragma: no-cache
Expires: Thu, 01 Jan 1970 00:00:00 GMT
Cache-Control: no-cache
Cache-Control: no-store
Content-Type: text/html;charset=UTF-8
Content-Language: en
Date: Tue, 27 Aug 2013 13:05:05 GMT
Connection: close

foo body
with multiples lines
"""
            }
        )
        crawler = get_crawler(SimpleSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                mockserver.url(f"/raw?{query}"), mockserver=mockserver
            )
        assert caplog.text.count("Got response 200") == 1

    @coroutine_test
    async def test_retry_conn_lost(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        # connection lost after receiving data
        crawler = get_crawler(SimpleSpider)
        with caplog.at_level(logging.DEBUG):
            await crawler.crawl_async(
                mockserver.url("/drop?abort=0"), mockserver=mockserver
            )
        self._assert_retried(caplog.text)

    @coroutine_test
    async def test_retry_conn_aborted(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        # connection lost before receiving data
        crawler = get_crawler(SimpleSpider)
        with caplog.at_level(logging.DEBUG):
            await crawler.crawl_async(
                mockserver.url("/drop?abort=1"), mockserver=mockserver
            )
        self._assert_retried(caplog.text)

    @staticmethod
    def _assert_retried(log: str) -> None:
        assert str(log).count("Retrying") == 2
        assert str(log).count("Gave up retrying") == 1

    @coroutine_test
    async def test_referer_header(self, mockserver: MockServer) -> None:
        """Referer header is set by RefererMiddleware unless it is already set"""
        req0 = Request(mockserver.url("/echo?headers=1&body=0"), dont_filter=True)
        req1 = req0.replace()
        req2 = req0.replace(headers={"Referer": None})
        req3 = req0.replace(headers={"Referer": "http://example.com"})
        req0.meta["next"] = req1
        req1.meta["next"] = req2
        req2.meta["next"] = req3
        crawler = get_crawler(SingleRequestSpider)
        await crawler.crawl_async(seed=req0, mockserver=mockserver)
        assert isinstance(crawler.spider, SingleRequestSpider)
        # basic asserts in case of weird communication errors
        assert "responses" in crawler.spider.meta
        assert "failures" not in crawler.spider.meta
        # start() doesn't set Referer header
        echo0 = json.loads(to_unicode(crawler.spider.meta["responses"][0].body))
        assert "Referer" not in echo0["headers"]
        # following request sets Referer to the source request url
        echo1 = json.loads(to_unicode(crawler.spider.meta["responses"][1].body))
        assert echo1["headers"].get("Referer") == [req0.url]
        # next request avoids Referer header
        echo2 = json.loads(to_unicode(crawler.spider.meta["responses"][2].body))
        assert "Referer" not in echo2["headers"]
        # last request explicitly sets a Referer header
        echo3 = json.loads(to_unicode(crawler.spider.meta["responses"][3].body))
        assert echo3["headers"].get("Referer") == ["http://example.com"]

    @coroutine_test
    async def test_engine_status(self, mockserver: MockServer) -> None:
        est: list[list[tuple[str, Any]]] = []

        def cb(response):
            est.append(get_engine_status(crawler.engine))

        crawler = get_crawler(SingleRequestSpider)
        await crawler.crawl_async(
            seed=mockserver.url("/"), callback_func=cb, mockserver=mockserver
        )
        assert isinstance(crawler.spider, SingleRequestSpider)
        assert len(est) == 1, est
        s = dict(est[0])
        assert s["engine.spider.name"] == crawler.spider.name
        assert s["len(engine.scraper.slot.active)"] == 1

    @coroutine_test
    async def test_print_engine_status(
        self, mockserver: MockServer, capsys: pytest.CaptureFixture[str]
    ) -> None:
        def cb(response):
            assert crawler.engine
            print_engine_status(crawler.engine)

        crawler = get_crawler(SingleRequestSpider)
        await crawler.crawl_async(
            seed=mockserver.url("/"), callback_func=cb, mockserver=mockserver
        )
        assert isinstance(crawler.spider, SingleRequestSpider)
        out = capsys.readouterr().out
        assert out.startswith("Execution engine status")
        assert re.search(rf"engine\.spider\.name +: {crawler.spider.name}\n", out)

    @coroutine_test
    async def test_format_engine_status(self, mockserver: MockServer) -> None:
        est: list[str] = []

        def cb(response):
            est.append(format_engine_status(crawler.engine))

        crawler = get_crawler(SingleRequestSpider)
        await crawler.crawl_async(
            seed=mockserver.url("/"), callback_func=cb, mockserver=mockserver
        )
        assert isinstance(crawler.spider, SingleRequestSpider)
        assert len(est) == 1, est
        est = est[0].split("\n")[2:-2]  # remove header & footer
        # convert to dict
        est_split = [x.split(":") for x in est]
        est = [x for sublist in est_split for x in sublist]  # flatten
        est = [x.lstrip().rstrip() for x in est]
        it = iter(est)
        s = dict(zip(it, it, strict=True))

        assert s["engine.spider.name"] == crawler.spider.name
        assert s["len(engine.scraper.slot.active)"] == "1"

    @coroutine_test
    async def test_open_spider_error_on_faulty_pipeline(
        self, mockserver: MockServer
    ) -> None:
        settings = {
            "ITEM_PIPELINES": {
                "tests.pipelines.ZeroDivisionErrorPipeline": 300,
            }
        }
        crawler = get_crawler(SimpleSpider, settings)
        with pytest.raises(ZeroDivisionError):
            await crawler.crawl_async(
                mockserver.url("/status?n=200"), mockserver=mockserver
            )
        assert not crawler.crawling

    @coroutine_test
    async def test_open_spider_error_on_faulty_pipeline_crawl(
        self, mockserver: MockServer
    ) -> None:
        # cover the except block in Crawler.crawl()
        settings = {
            "ITEM_PIPELINES": {
                "tests.pipelines.ZeroDivisionErrorPipeline": 300,
            }
        }
        crawler = get_crawler(SimpleSpider, settings)
        with pytest.raises(ZeroDivisionError):
            await maybe_deferred_to_future(
                crawler.crawl(mockserver.url("/status?n=200"), mockserver=mockserver)
            )
        assert not crawler.crawling

    @coroutine_test
    async def test_unknown_url_scheme(self, caplog: pytest.LogCaptureFixture) -> None:
        crawler = get_crawler(SimpleSpider)
        await crawler.crawl_async("foo://bar")
        assert "NotSupported: Unsupported URL scheme 'foo'" in caplog.text


class TestCrawlSpider:
    @staticmethod
    async def _run_spider(
        spider_cls: type[Spider], mockserver: MockServer
    ) -> tuple[list[Any], StatsCollector]:
        items = []

        def _on_item_scraped(item):
            items.append(item)

        crawler = get_crawler(spider_cls)
        crawler.signals.connect(_on_item_scraped, signals.item_scraped)
        await crawler.crawl_async(
            mockserver.url("/status?n=200"), mockserver=mockserver
        )
        return items, crawler.stats

    @coroutine_test
    async def test_crawlspider_with_parse(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(CrawlSpiderWithParseMethod)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(mockserver=mockserver)

        assert "[parse] status 200 (foo: None)" in caplog.text
        assert "[parse] status 201 (foo: None)" in caplog.text
        assert "[parse] status 202 (foo: bar)" in caplog.text

    @coroutine_test
    async def test_crawlspider_with_async_callback(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(CrawlSpiderWithAsyncCallback)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(mockserver=mockserver)

        assert "[parse_async] status 200 (foo: None)" in caplog.text
        assert "[parse_async] status 201 (foo: None)" in caplog.text
        assert "[parse_async] status 202 (foo: bar)" in caplog.text

    @coroutine_test
    async def test_crawlspider_with_async_generator_callback(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(CrawlSpiderWithAsyncGeneratorCallback)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(mockserver=mockserver)

        assert "[parse_async_gen] status 200 (foo: None)" in caplog.text
        assert "[parse_async_gen] status 201 (foo: None)" in caplog.text
        assert "[parse_async_gen] status 202 (foo: bar)" in caplog.text

    @coroutine_test
    async def test_crawlspider_with_errback(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(CrawlSpiderWithErrback)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(mockserver=mockserver)

        assert "[parse] status 200 (foo: None)" in caplog.text
        assert "[parse] status 201 (foo: None)" in caplog.text
        assert "[parse] status 202 (foo: bar)" in caplog.text
        assert "[errback] status 404" in caplog.text
        assert "[errback] status 500" in caplog.text
        assert "[errback] status 501" in caplog.text

    @coroutine_test
    async def test_crawlspider_without_errback(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(CrawlSpiderWithoutErrback)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(mockserver=mockserver)

        # The failing request (404) is followed by a rule without an errback,
        # so the failure is dropped silently and the crawl finishes normally.
        assert "[parse] status 200 (foo: None)" in caplog.text
        assert "[errback]" not in caplog.text
        assert crawler.stats.get_value("downloader/response_status_count/404") == 1

    @coroutine_test
    async def test_crawlspider_process_request_cb_kwargs(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(CrawlSpiderWithProcessRequestCallbackKeywordArguments)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(mockserver=mockserver)

        assert "[parse] status 200 (foo: process_request)" in caplog.text
        assert "[parse] status 201 (foo: process_request)" in caplog.text
        assert "[parse] status 202 (foo: bar)" in caplog.text

    @coroutine_test
    async def test_async_def_parse(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(AsyncDefSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                mockserver.url("/status?n=200"), mockserver=mockserver
            )
        assert "Got response 200" in caplog.text

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_asyncio_parse(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(
            AsyncDefAsyncioSpider,
            {
                "TWISTED_REACTOR": "twisted.internet.asyncioreactor.AsyncioSelectorReactor"
            },
        )
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                mockserver.url("/status?n=200"), mockserver=mockserver
            )
        assert "Got response 200" in caplog.text

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_asyncio_parse_items_list(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        with caplog.at_level(logging.INFO):
            items, _ = await self._run_spider(AsyncDefAsyncioReturnSpider, mockserver)
        assert "Got response 200" in caplog.text
        assert {"id": 1} in items
        assert {"id": 2} in items

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_asyncio_parse_items_single_element(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        items = []

        def _on_item_scraped(item):
            items.append(item)

        crawler = get_crawler(AsyncDefAsyncioReturnSingleElementSpider)
        crawler.signals.connect(_on_item_scraped, signals.item_scraped)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                mockserver.url("/status?n=200"), mockserver=mockserver
            )
        assert "Got response 200" in caplog.text
        assert {"foo": 42} in items

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_asyncgen_parse(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        with caplog.at_level(logging.INFO):
            _, stats = await self._run_spider(AsyncDefAsyncioGenSpider, mockserver)
        assert "Got response 200" in caplog.text
        itemcount = stats.get_value("item_scraped_count")
        assert itemcount == 1

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_asyncgen_parse_loop(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        with caplog.at_level(logging.INFO):
            items, stats = await self._run_spider(
                AsyncDefAsyncioGenLoopSpider, mockserver
            )
        assert "Got response 200" in caplog.text
        itemcount = stats.get_value("item_scraped_count")
        assert itemcount == 10
        for i in range(10):
            assert {"foo": i} in items

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_asyncgen_parse_exc(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        with caplog.at_level(logging.INFO):
            items, stats = await self._run_spider(
                AsyncDefAsyncioGenExcSpider, mockserver
            )
        assert "Spider error processing" in caplog.text
        assert "ValueError" in caplog.text
        itemcount = stats.get_value("item_scraped_count")
        assert itemcount == 7
        for i in range(7):
            assert {"foo": i} in items

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_asyncgen_parse_complex(
        self, mockserver: MockServer
    ) -> None:
        items, stats = await self._run_spider(
            AsyncDefAsyncioGenComplexSpider, mockserver
        )
        itemcount = stats.get_value("item_scraped_count")
        assert itemcount == 156
        # some random items
        for i in [1, 4, 21, 22, 207, 311]:
            assert {"index": i} in items
        for i in [10, 30, 122]:
            assert {"index2": i} in items

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_asyncio_parse_reqs_list(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        with caplog.at_level(logging.INFO):
            await self._run_spider(AsyncDefAsyncioReqsReturnSpider, mockserver)
        for req_id in range(3):
            assert f"Got response 200, req_id {req_id}" in caplog.text

    @pytest.mark.only_not_asyncio
    @coroutine_test
    async def test_async_def_deferred_direct(self, mockserver: MockServer) -> None:
        items, _ = await self._run_spider(AsyncDefDeferredDirectSpider, mockserver)
        assert items == [{"code": 200}]

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_async_def_deferred_wrapped(self, mockserver: MockServer) -> None:
        items, _ = await self._run_spider(AsyncDefDeferredWrappedSpider, mockserver)
        assert items == [{"code": 200}]

    @coroutine_test
    async def test_async_def_deferred_maybe_wrapped(
        self, mockserver: MockServer
    ) -> None:
        items, _ = await self._run_spider(
            AsyncDefDeferredMaybeWrappedSpider, mockserver
        )
        assert items == [{"code": 200}]

    @coroutine_test
    async def test_response_ssl_certificate_none(self, mockserver: MockServer) -> None:
        crawler = get_crawler(SingleRequestSpider)
        url = mockserver.url("/echo?body=test", is_secure=False)
        await crawler.crawl_async(seed=url, mockserver=mockserver)
        assert isinstance(crawler.spider, SingleRequestSpider)
        assert crawler.spider.meta["responses"][0].certificate is None

    @pytest.mark.filterwarnings(
        r"ignore:.*You should use cryptography's X\.509 APIs:DeprecationWarning"
    )
    @pytest.mark.parametrize(
        "url",
        [
            "/echo?body=test",
            pytest.param(
                "/status?n=200",
                marks=pytest.mark.xfail(
                    'config.getoption("--reactor") != "none"',
                    reason="With HTTP11DownloadHandler, responses with no body are returned early and contain no certificate",
                    strict=True,
                ),
            ),
        ],
    )
    @coroutine_test
    async def test_response_ssl_certificate(
        self, mockserver: MockServer, url: str
    ) -> None:
        crawler = get_crawler(SingleRequestSpider)
        url = mockserver.url(url, is_secure=True)
        await crawler.crawl_async(seed=url, mockserver=mockserver)
        assert isinstance(crawler.spider, SingleRequestSpider)
        cert = crawler.spider.meta["responses"][0].certificate
        assert cert is not None
        if isinstance(cert, Certificate):  # Twisted
            assert cert.getSubject().commonName == b"localhost"
            assert cert.getIssuer().commonName == b"localhost"
        elif isinstance(cert, bytes):  # DER bytes
            cert_x509 = load_der_x509_certificate(cert)
            assert cert_x509.subject.rfc4514_string() == "CN=localhost,O=Scrapy,C=IE"
            assert cert_x509.issuer.rfc4514_string() == "CN=localhost,O=Scrapy,C=IE"

    @pytest.mark.parametrize(
        "url",
        [
            "/echo?body=test",
            pytest.param(
                "/status?n=200",
                marks=pytest.mark.xfail(
                    'config.getoption("--reactor") != "none"',
                    reason="With HTTP11DownloadHandler, responses with no body are returned early and contain no ip_address",
                    strict=True,
                ),
            ),
        ],
    )
    @coroutine_test
    async def test_response_ip_address(self, mockserver: MockServer, url: str) -> None:
        crawler = get_crawler(SingleRequestSpider)
        url = mockserver.url(url)
        expected_netloc, _ = urlparse(url).netloc.split(":")
        await crawler.crawl_async(seed=url, mockserver=mockserver)
        assert isinstance(crawler.spider, SingleRequestSpider)
        ip_address = crawler.spider.meta["responses"][0].ip_address
        assert isinstance(ip_address, IPv4Address)
        assert str(ip_address) == gethostbyname(expected_netloc)

    @coroutine_test
    async def test_bytes_received_stop_download_callback(
        self, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(BytesReceivedCallbackSpider)
        await crawler.crawl_async(mockserver=mockserver)
        assert isinstance(crawler.spider, BytesReceivedCallbackSpider)
        assert crawler.spider.meta.get("failure") is None
        assert isinstance(crawler.spider.meta["response"], Response)
        assert crawler.spider.meta["response"].body == crawler.spider.meta.get(
            "bytes_received"
        )
        assert (
            len(crawler.spider.meta["response"].body)
            < crawler.spider.full_response_length
        )

    @coroutine_test
    async def test_bytes_received_stop_download_errback(
        self, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(BytesReceivedErrbackSpider)
        await crawler.crawl_async(mockserver=mockserver)
        assert isinstance(crawler.spider, BytesReceivedErrbackSpider)
        assert crawler.spider.meta.get("response") is None
        assert isinstance(crawler.spider.meta["failure"], Failure)
        assert isinstance(crawler.spider.meta["failure"].value, StopDownload)
        assert isinstance(crawler.spider.meta["failure"].value.response, Response)
        assert crawler.spider.meta[
            "failure"
        ].value.response.body == crawler.spider.meta.get("bytes_received")
        assert (
            len(crawler.spider.meta["failure"].value.response.body)
            < crawler.spider.full_response_length
        )

    @coroutine_test
    async def test_headers_received_stop_download_callback(
        self, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(HeadersReceivedCallbackSpider)
        await crawler.crawl_async(mockserver=mockserver)
        assert isinstance(crawler.spider, HeadersReceivedCallbackSpider)
        assert crawler.spider.meta.get("failure") is None
        assert isinstance(crawler.spider.meta["response"], Response)
        assert crawler.spider.meta["response"].headers == crawler.spider.meta.get(
            "headers_received"
        )

    @coroutine_test
    async def test_headers_received_stop_download_errback(
        self, mockserver: MockServer
    ) -> None:
        crawler = get_crawler(HeadersReceivedErrbackSpider)
        await crawler.crawl_async(mockserver=mockserver)
        assert isinstance(crawler.spider, HeadersReceivedErrbackSpider)
        assert crawler.spider.meta.get("response") is None
        assert isinstance(crawler.spider.meta["failure"], Failure)
        assert isinstance(crawler.spider.meta["failure"].value, StopDownload)
        assert isinstance(crawler.spider.meta["failure"].value.response, Response)
        assert crawler.spider.meta[
            "failure"
        ].value.response.headers == crawler.spider.meta.get("headers_received")

    @coroutine_test
    async def test_spider_callback_deferred_deprecated(
        self, mockserver: MockServer
    ) -> None:
        def cb(response: Response) -> Any:
            return succeed(None)

        crawler = get_crawler(SingleRequestSpider)
        with pytest.warns(
            ScrapyDeprecationWarning,
            match="Returning Deferreds from spider callbacks is deprecated",
        ):
            await crawler.crawl_async(seed=mockserver.url("/"), callback_func=cb)

    @coroutine_test
    async def test_spider_errback(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        failures = []

        def eb(failure: Failure) -> Failure:
            failures.append(failure)
            return failure

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                seed=mockserver.url("/status?n=400"), errback_func=eb
            )
        assert len(failures) == 1
        assert "HTTP status code is not handled or not allowed" in caplog.text
        assert "Spider error processing" not in caplog.text

    @coroutine_test
    async def test_spider_errback_silence(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        failures = []

        def eb(failure: Failure) -> None:
            failures.append(failure)

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                seed=mockserver.url("/status?n=400"), errback_func=eb
            )
        assert len(failures) == 1
        assert "HTTP status code is not handled or not allowed" not in caplog.text
        assert "Spider error processing" not in caplog.text

    @coroutine_test
    async def test_spider_errback_exception(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        def eb(failure: Failure) -> None:
            raise ValueError("foo")

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                seed=mockserver.url("/status?n=400"), errback_func=eb
            )
        assert "Spider error processing" in caplog.text

    @coroutine_test
    async def test_spider_errback_item(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        def eb(failure: Failure) -> Any:
            return {"foo": "bar"}

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                seed=mockserver.url("/status?n=400"), errback_func=eb
            )
        assert "HTTP status code is not handled or not allowed" not in caplog.text
        assert "Spider error processing" not in caplog.text
        assert "'item_scraped_count': 1" in caplog.text

    @coroutine_test
    async def test_spider_errback_request(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        def eb(failure: Failure) -> Request:
            return Request(mockserver.url("/"))

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.DEBUG):
            await crawler.crawl_async(
                seed=mockserver.url("/status?n=400"), errback_func=eb
            )
        assert "HTTP status code is not handled or not allowed" not in caplog.text
        assert "Spider error processing" not in caplog.text
        assert "Crawled (200)" in caplog.text

    @coroutine_test
    async def test_spider_errback_downloader_error(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        failures = []

        def eb(failure: Failure) -> Failure:
            failures.append(failure)
            return failure

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                seed=mockserver.url("/drop?abort=1"), errback_func=eb
            )
        assert len(failures) == 1
        assert "Error downloading" in caplog.text
        assert "Spider error processing" not in caplog.text

    @coroutine_test
    async def test_spider_errback_downloader_error_exception(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        def eb(failure: Failure) -> None:
            raise ValueError("foo")

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                seed=mockserver.url("/drop?abort=1"), errback_func=eb
            )
        assert "Error downloading" in caplog.text
        assert "Spider error processing" in caplog.text

    @coroutine_test
    async def test_spider_errback_downloader_error_item(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        def eb(failure: Failure) -> Any:
            return {"foo": "bar"}

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(
                seed=mockserver.url("/drop?abort=1"), errback_func=eb
            )
        assert "HTTP status code is not handled or not allowed" not in caplog.text
        assert "Spider error processing" not in caplog.text
        assert "'item_scraped_count': 1" in caplog.text

    @coroutine_test
    async def test_spider_errback_downloader_error_request(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        def eb(failure: Failure) -> Request:
            return Request(mockserver.url("/"))

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.DEBUG):
            await crawler.crawl_async(
                seed=mockserver.url("/drop?abort=1"), errback_func=eb
            )
        assert "HTTP status code is not handled or not allowed" not in caplog.text
        assert "Spider error processing" not in caplog.text
        assert "Crawled (200)" in caplog.text

    @coroutine_test
    async def test_spider_errback_deferred_deprecated(
        self, mockserver: MockServer
    ) -> None:
        def eb(failure: Failure) -> Any:
            return succeed(None)

        crawler = get_crawler(SingleRequestSpider)
        with pytest.warns(
            ScrapyDeprecationWarning,
            match="Returning Deferreds from spider errbacks is deprecated",
        ):
            await crawler.crawl_async(
                seed=mockserver.url("/status?n=400"), errback_func=eb
            )

    @coroutine_test
    async def test_raise_closespider(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        def cb(response):
            raise CloseSpider

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(seed=mockserver.url("/"), callback_func=cb)
        assert "Closing spider (cancelled)" in caplog.text
        assert "Spider error processing" not in caplog.text

    @coroutine_test
    async def test_raise_closespider_reason(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        def cb(response):
            raise CloseSpider("my_reason")

        crawler = get_crawler(SingleRequestSpider)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async(seed=mockserver.url("/"), callback_func=cb)
        assert "Closing spider (my_reason)" in caplog.text
        assert "Spider error processing" not in caplog.text



"""Helper functions which don't fit anywhere else"""

from __future__ import annotations

import ast
import inspect
import os
import re
import warnings
from collections import deque
from contextlib import contextmanager
from functools import partial
from importlib import import_module
from pkgutil import iter_modules
from typing import TYPE_CHECKING, Any, ParamSpec, Protocol, TypeVar, cast, overload

from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.item import Item
from scrapy.utils.datatypes import LocalWeakReferencedCache

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator
    from types import ModuleType

    from scrapy import Spider
    from scrapy.crawler import Crawler


_ITERABLE_SINGLE_VALUES = dict, Item, str, bytes
_ITER_T = TypeVar("_ITER_T", bound=dict[Any, Any] | Item | str | bytes)
_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_P = ParamSpec("_P")


@overload
def arg_to_iter(arg: None) -> tuple[()]: ...
@overload
def arg_to_iter(arg: _ITER_T) -> Iterable[_ITER_T]: ...
@overload
def arg_to_iter(arg: Iterable[_T]) -> Iterable[_T]: ...
@overload
def arg_to_iter(arg: _T) -> Iterable[_T]: ...
def arg_to_iter(arg: Any) -> Iterable[Any]:
    """Convert an argument to an iterable. The argument can be a None, single
    value, or an iterable.

    Exception: if arg is a dict, [arg] will be returned
    """
    if arg is None:
        return ()
    if not isinstance(arg, _ITERABLE_SINGLE_VALUES) and hasattr(arg, "__iter__"):
        return cast("Iterable[Any]", arg)
    return [arg]


def load_object(path: str | Callable[..., Any]) -> Any:
    """Load an object given its absolute object path, and return it.

    The object can be the import path of a class, function, variable or an
    instance, e.g. 'scrapy.downloadermiddlewares.redirect.RedirectMiddleware'.

    If ``path`` is not a string, but is a callable object, such as a class or
    a function, then return it as is.
    """

    if not isinstance(path, str):
        if callable(path):
            return path
        raise TypeError(
            f"Unexpected argument type, expected string or object, got: {type(path)}"
        )

    try:
        dot = path.rindex(".")
    except ValueError:
        raise ValueError(f"Error loading object '{path}': not a full path") from None

    module, name = path[:dot], path[dot + 1 :]
    mod = import_module(module)

    try:
        obj = getattr(mod, name)
    except AttributeError:
        raise NameError(
            f"Module '{module}' doesn't define any object named '{name}'"
        ) from None

    return obj


def _load_objects(objects: Iterable[str | Callable[..., Any]]) -> tuple[Any, ...]:
    """Resolve *objects* (objects or import paths) to a tuple of objects."""
    return tuple(load_object(obj) if isinstance(obj, str) else obj for obj in objects)


def walk_modules_iter(path: str) -> Iterable[ModuleType]:
    """Loads a module and all its submodules from the given module path and
    returns them. If *any* module throws an exception while importing, that
    exception is thrown back.

    For example:
    >>> list(walk_modules_iter('scrapy.utils'))
    [<module 'scrapy.utils' from '...'>, ...]
    >>> gen = walk_modules_iter('scrapy.utils.nonexistent') # error not raised until the generator is consumed
    >>> list(gen)
    Traceback (most recent call last):
        ...
    ModuleNotFoundError: No module named 'scrapy.utils.nonexistent'...
    """

    mod = import_module(path)
    yield mod
    if hasattr(mod, "__path__"):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + "." + subpath
            if ispkg:
                yield from walk_modules_iter(fullpath)
            else:
                yield import_module(fullpath)


def walk_modules(path: str) -> list[ModuleType]:  # pragma: no cover
    """
    Loads a module and all its submodules from the given module path and
    returns them. If *any* module throws an exception while importing, that
    exception is thrown back.
    """
    warnings.warn(
        (
            "The scrapy.utils.misc.walk_modules function is deprecated and will be "
            "removed in a future version of Scrapy. "
            "Use scrapy.utils.misc.walk_modules_iter instead."
        ),
        ScrapyDeprecationWarning,
        stacklevel=2,
    )

    return list(walk_modules_iter(path))


def rel_has_nofollow(rel: str | None) -> bool:
    """Return True if link rel attribute has nofollow type"""
    return rel is not None and "nofollow" in rel.lower().replace(",", " ").split()


class SupportsFromCrawler(Protocol[_T_co, _P]):
    @classmethod
    def from_crawler(
        cls, crawler: Crawler, /, *args: _P.args, **kwargs: _P.kwargs
    ) -> _T_co: ...


@overload
def build_from_crawler(
    objcls: SupportsFromCrawler[_T_co, _P],
    crawler: Crawler,
    /,
    *args: _P.args,
    **kwargs: _P.kwargs,
) -> _T_co: ...


@overload
def build_from_crawler(
    objcls: Callable[_P, _T_co],
    crawler: Crawler,
    /,
    *args: _P.args,
    **kwargs: _P.kwargs,
) -> _T_co: ...


def build_from_crawler(
    objcls: Any,
    crawler: Crawler,
    /,
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Construct a class instance using its ``from_crawler()`` or ``__init__()`` constructor.

    .. versionadded:: 2.12

    ``*args`` and ``**kwargs`` are forwarded to the constructor.

    Raises ``TypeError`` if the resulting instance is ``None``.
    """
    if hasattr(objcls, "from_crawler"):
        instance = objcls.from_crawler(crawler, *args, **kwargs)
        method_name = "from_crawler"
    else:
        instance = objcls(*args, **kwargs)
        method_name = "__new__"
    if instance is None:
        raise TypeError(f"{objcls.__qualname__}.{method_name} returned None")
    return instance


@contextmanager
def set_environ(**kwargs: str) -> Iterator[None]:
    """Temporarily set environment variables inside the context manager and
    fully restore previous environment afterwards
    """

    original_env = {k: os.environ.get(k) for k in kwargs}
    os.environ.update(kwargs)
    try:
        yield
    finally:
        for k, v in original_env.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v


def walk_callable(node: ast.AST) -> Iterable[ast.AST]:
    """Similar to ``ast.walk``, but walks only function body and skips nested
    functions defined within the node.
    """
    todo: deque[ast.AST] = deque([node])
    walked_func_def = False
    while todo:
        node = todo.popleft()
        if isinstance(node, ast.FunctionDef):
            if walked_func_def:
                continue
            walked_func_def = True
        todo.extend(ast.iter_child_nodes(node))
        yield node


_generator_callbacks_cache: LocalWeakReferencedCache[Callable[..., Any], bool] = (
    LocalWeakReferencedCache(limit=128)
)


def _returns_none(return_node: ast.Return) -> bool:
    value = return_node.value
    return value is None or (isinstance(value, ast.Constant) and value.value is None)


def is_generator_with_return_value(callable: Callable[..., Any]) -> bool:  # noqa: A002
    """
    Returns True if a callable is a generator function which includes a
    'return' statement with a value different than None, False otherwise
    """
    if callable in _generator_callbacks_cache:
        return bool(_generator_callbacks_cache[callable])

    if inspect.isgeneratorfunction(callable):
        func = callable
        while isinstance(func, partial):
            func = func.func

        src = inspect.getsource(func)
        pattern = re.compile(r"(^[\t ]+)")
        code = pattern.sub("", src)

        match = pattern.match(src)  # finds indentation
        if match:
            code = re.sub(f"\n{match.group(0)}", "\n", code)  # remove indentation

        tree = ast.parse(code)
        for node in walk_callable(tree):
            if isinstance(node, ast.Return) and not _returns_none(node):
                _generator_callbacks_cache[callable] = True
                return bool(_generator_callbacks_cache[callable])

    _generator_callbacks_cache[callable] = False
    return bool(_generator_callbacks_cache[callable])


def warn_on_generator_with_return_value(
    spider: Spider,
    callable: Callable[..., Any],  # noqa: A002
) -> None:
    """
    Logs a warning if a callable is a generator function and includes
    a 'return' statement with a value different than None
    """
    if not spider.settings.getbool("WARN_ON_GENERATOR_RETURN_VALUE"):
        return
    try:
        if is_generator_with_return_value(callable):
            warnings.warn(
                f'The "{spider.__class__.__name__}.{callable.__name__}" method is '
                'a generator and includes a "return" statement with a value '
                "different than None. This could lead to unexpected behaviour. Please see "
                "https://docs.python.org/3/reference/simple_stmts.html#the-return-statement "
                'for details about the semantics of the "return" statement within generators',
                stacklevel=2,
            )
    except IndentationError:
        callable_name = spider.__class__.__name__ + "." + callable.__name__
        warnings.warn(
            f'Unable to determine whether or not "{callable_name}" is a generator with a return value. '
            "This will not prevent your code from working, but it prevents Scrapy from detecting "
            f'potential issues in your implementation of "{callable_name}". Please, report this in the '
            "Scrapy issue tracker (https://github.com/scrapy/scrapy/issues), "
            f'including the code of "{callable_name}"',
            stacklevel=2,
        )



from __future__ import annotations

import os
import sys
from pathlib import Path
from unittest import mock

import pytest

from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.item import Field, Item
from scrapy.utils.misc import (
    arg_to_iter,
    build_from_crawler,
    load_object,
    rel_has_nofollow,
    set_environ,
    walk_modules,
    walk_modules_iter,
)


def test_load_object_class() -> None:
    obj = load_object(Field)
    assert obj is Field
    obj = load_object("scrapy.item.Field")
    assert obj is Field


def test_load_object_function() -> None:
    obj = load_object(load_object)
    assert obj is load_object
    obj = load_object("scrapy.utils.misc.load_object")
    assert obj is load_object


def test_load_object_exceptions() -> None:
    with pytest.raises(ImportError):
        load_object("nomodule999.mod.function")
    with pytest.raises(NameError):
        load_object("scrapy.utils.misc.load_object999")
    with pytest.raises(TypeError):
        load_object({})  # type: ignore[arg-type]


def test_walk_modules() -> None:
    mods = walk_modules_iter("tests.test_utils_misc.test_walk_modules")
    expected = [
        "tests.test_utils_misc.test_walk_modules",
        "tests.test_utils_misc.test_walk_modules.mod",
        "tests.test_utils_misc.test_walk_modules.mod.mod0",
        "tests.test_utils_misc.test_walk_modules.mod1",
    ]
    assert {m.__name__ for m in mods} == set(expected)

    mods = walk_modules_iter("tests.test_utils_misc.test_walk_modules.mod")
    expected = [
        "tests.test_utils_misc.test_walk_modules.mod",
        "tests.test_utils_misc.test_walk_modules.mod.mod0",
    ]
    assert {m.__name__ for m in mods} == set(expected)

    mods = walk_modules_iter("tests.test_utils_misc.test_walk_modules.mod1")
    expected = [
        "tests.test_utils_misc.test_walk_modules.mod1",
    ]
    assert {m.__name__ for m in mods} == set(expected)

    with pytest.raises(ImportError):
        for _ in walk_modules_iter("nomodule999"):
            pass
    with (
        pytest.raises(ImportError),
        pytest.warns(
            ScrapyDeprecationWarning,
            match="The scrapy.utils.misc.walk_modules function is deprecated and will be "
            "removed in a future version of Scrapy. "
            "Use scrapy.utils.misc.walk_modules_iter instead.",
        ),
    ):
        walk_modules("nomodule999")


def test_walk_modules_egg() -> None:
    egg = str(Path(__file__).parent / "test.egg")
    sys.path.append(egg)
    try:
        mods = walk_modules_iter("testegg")
        expected = [
            "testegg.spiders",
            "testegg.spiders.a",
            "testegg.spiders.b",
            "testegg",
        ]
        assert {m.__name__ for m in mods} == set(expected)
    finally:
        sys.path.remove(egg)


def test_arg_to_iter() -> None:
    class TestItem(Item):
        name = Field()

    assert hasattr(arg_to_iter(None), "__iter__")
    assert hasattr(arg_to_iter(100), "__iter__")
    assert hasattr(arg_to_iter("lala"), "__iter__")
    assert hasattr(arg_to_iter([1, 2, 3]), "__iter__")
    assert hasattr(arg_to_iter(c for c in "abcd"), "__iter__")

    assert not list(arg_to_iter(None))
    assert list(arg_to_iter("lala")) == ["lala"]
    assert list(arg_to_iter(100)) == [100]
    assert list(arg_to_iter(c for c in "abc")) == ["a", "b", "c"]
    assert list(arg_to_iter([1, 2, 3])) == [1, 2, 3]
    assert list(arg_to_iter({"a": 1})) == [{"a": 1}]
    assert list(arg_to_iter(TestItem(name="john"))) == [TestItem(name="john")]


def test_build_from_crawler() -> None:
    crawler = mock.MagicMock(spec_set=["settings"])
    args = (True, 100.0)
    kwargs = {"key": "val"}

    def _test_with_crawler(mock: mock.MagicMock, crawler: mock.MagicMock) -> None:
        build_from_crawler(mock, crawler, *args, **kwargs)
        if hasattr(mock, "from_crawler"):
            mock.from_crawler.assert_called_once_with(crawler, *args, **kwargs)
            assert mock.call_count == 0
        else:
            mock.assert_called_once_with(*args, **kwargs)

    # Check usage of correct constructor using 2 mocks:
    #   1. with no alternative constructors
    #   2. with from_crawler() constructor
    spec_sets = (
        ["__qualname__"],
        ["__qualname__", "from_crawler"],
    )
    for specs in spec_sets:
        m = mock.MagicMock(spec_set=specs)
        _test_with_crawler(m, crawler)
        m.reset_mock()

    # Check adoption of crawler
    m = mock.MagicMock(spec_set=["__qualname__", "from_crawler"])
    m.from_crawler.return_value = None
    with pytest.raises(TypeError):
        build_from_crawler(m, crawler, *args, **kwargs)


def test_set_environ() -> None:
    assert os.environ.get("some_test_environ") is None
    with set_environ(some_test_environ="test_value"):
        assert os.environ.get("some_test_environ") == "test_value"
    assert os.environ.get("some_test_environ") is None

    os.environ["some_test_environ"] = "test"
    assert os.environ.get("some_test_environ") == "test"
    with set_environ(some_test_environ="test_value"):
        assert os.environ.get("some_test_environ") == "test_value"
    assert os.environ.get("some_test_environ") == "test"


def test_rel_has_nofollow() -> None:
    assert rel_has_nofollow("ugc nofollow") is True
    assert rel_has_nofollow("ugc,nofollow") is True
    assert rel_has_nofollow("ugc") is False
    assert rel_has_nofollow("nofollow") is True
    assert rel_has_nofollow("nofollowfoo") is False
    assert rel_has_nofollow("foonofollow") is False
    assert rel_has_nofollow("ugc,  ,  nofollow") is True
    # rel attribute values are ASCII case-insensitive per the HTML spec
    assert rel_has_nofollow("NoFollow") is True
    assert rel_has_nofollow("NOFOLLOW") is True
    assert rel_has_nofollow("UGC NoFollow") is True
    assert rel_has_nofollow("ugc,NoFollow") is True



from __future__ import annotations

import pytest

from scrapy.downloadermiddlewares.stats import DownloaderStats, get_header_size
from scrapy.exceptions import NotConfigured
from scrapy.http import Request, Response
from scrapy.spiders import Spider
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler


class MyException(Exception):
    pass


class TestDownloaderStats:
    def setup_method(self) -> None:
        self.crawler = get_crawler(Spider)
        self.mw = build_from_crawler(DownloaderStats, self.crawler)

        self.crawler.stats.open_spider()

        self.req = Request("http://scrapytest.org")
        self.res = Response("http://scrapytest.org", status=400)

    def teardown_method(self) -> None:
        self.crawler.stats.close_spider()

    def assertStatsEqual(self, key: str, value: object) -> None:
        assert self.crawler.stats.get_value(key) == value, str(
            self.crawler.stats.get_stats()
        )

    def test_process_request(self) -> None:
        self.mw.process_request(self.req)
        self.assertStatsEqual("downloader/request_count", 1)

    def test_process_response(self) -> None:
        self.mw.process_response(self.req, self.res)
        self.assertStatsEqual("downloader/response_count", 1)

    def test_process_exception(self) -> None:
        self.mw.process_exception(self.req, MyException())
        self.assertStatsEqual("downloader/exception_count", 1)
        self.assertStatsEqual(
            "downloader/exception_type_count/tests.test_downloadermiddleware_stats.MyException",
            1,
        )

    def test_from_crawler_not_configured(self) -> None:
        crawler = get_crawler(Spider, {"DOWNLOADER_STATS": False})
        with pytest.raises(NotConfigured):
            build_from_crawler(DownloaderStats, crawler)


def test_get_header_size_non_list_value() -> None:
    # Deliberately passing a non-list/tuple header value to make sure
    # get_header_size() degrades gracefully instead of raising.
    assert get_header_size({"Content-Type": "text/html"}) == 0  # type: ignore[dict-item]
